"""

    thorn.django.managers
    =====================

    Managers and query sets.

"""
from __future__ import absolute_import, unicode_literals

from django.db import models
from django.db.models.query import Q

__all__ = ['SubscriberQuerySet', 'SubscriberManager']


class SubscriberQuerySet(models.QuerySet):

    def matching(self, event, user=None):
        return self.matching_event(event).matching_user_or_all(user)

    def matching_event(self, event):
        topic, _, rest = event.partition('.')
        # order.completed
        # order.*
        # *.completed
        return self.filter(
            Q(event__exact=event) |
            Q(event__exact='{0}.*'.format(topic)) |
            Q(event__exact='*.{0}'.format(rest))
        )

    def matching_user_or_all(self, user):
        return self.filter(user=user) if user else self


class SubscriberManager(models.Manager.from_queryset(SubscriberQuerySet)):
    pass



"""

    thorn.utils.log
    ===============

    Logging utilities.

"""
from __future__ import absolute_import, unicode_literals

from six import string_types

from kombu.log import get_logger as _get_logger

__all__ = ['base_logger', 'get_logger']

base_logger = _get_logger('webhooks')


def get_logger(name, parent=base_logger):
    assert isinstance(name, string_types)
    logger = _get_logger(name)
    logger.parent = parent
    return logger



"""

    thorn.tasks
    ===========

    Tasks used by the Celery dispatcher.

"""
from __future__ import absolute_import, unicode_literals

from celery import shared_task
from celery.utils.functional import memoize

from ._state import app_or_default

__all__ = ['send_event', 'dispatch_requests', 'dispatch_request']


@memoize()
def _worker_dispatcher():
    from .dispatch.celery import WorkerDispatcher
    return WorkerDispatcher()


@shared_task(ignore_result=True)
def send_event(event, payload, sender, timeout):
    """Task called by process dispatching the event.

    This will use the WorkerDispatcher to dispatch the individual
    HTTP requests in batches (``dispatch_requests -> dispatch_request``).

    """
    _worker_dispatcher().send(event, payload, sender, timeout=timeout)


@shared_task(ignore_result=True)
def dispatch_requests(reqs, app=None):
    """Process a batch of HTTP requests."""
    app = app_or_default(app)
    session = app.Request.Session()
    [dispatch_request(session=session, app=app, **req) for req in reqs]


@shared_task(bind=True, ignore_result=True)
def dispatch_request(self, event, data, sender, subscriber,
                     session=None, app=None, **kwargs):
    """Process a single HTTP request."""
    app = app_or_default(app)
    # the user is serialized as the pk, so we cannot pass it
    # directly to Subscriber, but we also don't need it at this point.
    subscriber.pop('user', None)
    subscriber = app.Subscriber(**subscriber)
    request = app.Request(event, data, sender, subscriber, **kwargs)
    try:
        request.dispatch(session=session, propagate=request.retry)
    except request.connection_errors + request.timeout_errors as exc:
        if request.retry:
            raise self.retry(exc=exc, max_retries=request.retry_max,
                             countdown=request.retry_delay)
        raise



"""

    thorn.conf
    ==========

    Webhooks-related configuration settings.

"""
from __future__ import absolute_import, unicode_literals

from celery.utils import cached_property

from . import validators
from ._state import app_or_default
from .exceptions import ImproperlyConfigured
from .utils import json

__all__ = ['settings', 'event_choices']

MIME_JSON = 'application/json'
MIME_URLFORM = 'application/x-www-form-urlencoded'


class Settings(object):

    app = None

    default_chunksize = 10
    default_dispatcher = 'default'
    default_event_choices = ()
    default_timeout = 3.0
    default_codecs = {MIME_JSON: json.dumps}
    default_drf_permission_classes = None
    default_retry = True
    default_retry_max = 10
    default_retry_delay = 60.0
    default_recipient_validators = [
        validators.block_internal_ips(),
        validators.ensure_protocol('http', 'https'),
        validators.ensure_port(80, 443),
    ]

    def __init__(self, app=None):
        self.app = app_or_default(app or self.app)

    @cached_property
    def THORN_CHUNKSIZE(self):
        return (
            getattr(self.app.config, 'THORN_CHUNKSIZE', None) or
            self.default_chunksize
        )

    @cached_property
    def THORN_CODECS(self):
        return (
            getattr(self.app.config, 'THORN_CODECS', None) or
            self.default_codecs
        )

    @cached_property
    def THORN_SUBSCRIBERS(self):
        return (
            getattr(self.app.config, 'THORN_SUBSCRIBERS', None) or {}
        )

    @cached_property
    def THORN_DISPATCHER(self):
        return (
            getattr(self.app.config, 'THORN_DISPATCHER', None) or
            self.default_dispatcher
        )

    @cached_property
    def THORN_EVENT_CHOICES(self):
        return (
            getattr(self.app.config, 'THORN_EVENT_CHOICES', None) or
            self.default_event_choices
        )

    @cached_property
    def THORN_DRF_PERMISSION_CLASSES(self):
        return (
            getattr(self.app.config, 'THORN_DRF_PERMISSION_CLASSES', None) or
            self.default_drf_permission_classes
        )

    @cached_property
    def THORN_EVENT_TIMEOUT(self):
        return (
            getattr(self.app.config, 'THORN_EVENT_TIMEOUT', None) or
            self.default_timeout
        )

    @cached_property
    def THORN_RETRY(self):
        return (
            getattr(self.app.config, 'THORN_RETRY', None) or
            self.default_retry
        )

    @cached_property
    def THORN_RETRY_MAX(self):
        return (
            getattr(self.app.config, 'THORN_RETRY_MAX', None) or
            self.default_retry_max
        )

    @cached_property
    def THORN_RETRY_DELAY(self):
        return (
            getattr(self.app.config, 'THORN_RETRY_DELAY', None) or
            self.default_retry_delay,
        )

    @cached_property
    def THORN_RECIPIENT_VALIDATORS(self):
        return getattr(
            self.app.config, 'THORN_RECIPIENT_VALIDATORS',
            self.default_recipient_validators,
        )
settings = Settings()


def event_choices(app=None):
    app = app_or_default(app)
    choices = app.settings.THORN_EVENT_CHOICES
    try:
        return list(zip(choices, choices))
    except TypeError:
        raise ImproperlyConfigured('THORN_EVENT_CHOICES not a list/tuple.')


def all_settings():
    return {n for n in dir(Settings) if n.isupper() and not n.startswith('__')}



from __future__ import absolute_import, unicode_literals

from thorn.utils.functional import Q, groupbymax, traverse_subscribers

from thorn.tests.case import Case, Mock, patch


class test_groupbymax(Case):

    def test_scalar(self):
        self.assertListEqual(
            list(groupbymax(['a'], 10)), [['a']],
        )

    def test_maxsize(self):
        self.assertListEqual(
            list(groupbymax('aaaabcdde', 3)),
            [['a', 'a', 'a'], ['a'], ['b'], ['c'], ['d', 'd'], ['e']],
        )

    def test_random(self):
        self.assertListEqual(
            list(groupbymax(range(10), 10)),
            [[i] for i in range(10)],
        )

    def test_buf_extends(self):
        self.assertListEqual(
            list(groupbymax('a' * 100, 200)),
            [['a'] * 100],
        )


class test_Q(Case):

    def test_missing_op(self):
        with self.assertRaises(ValueError):
            Q(foo=30)(Mock())

    def test_missing_op__nested(self):
        with self.assertRaises(ValueError):
            Q(foo__bar__baz=30)(Mock())

    def test_missing_attr(self):
        with self.assertRaises(AttributeError):
            Q(foo__eq=30)(object())

    def test_eq(self):
        self.assertTrue(Q(foo__eq=30)(Mock(foo=30)))

        self.assertFalse(Q(foo__eq=30)(Mock(foo=10)))
        self.assertFalse(Q(foo__eq=30)(Mock(foo=None)))

    def test_eq__True(self):
        self.assertTrue(Q(foo__eq=True)(Mock(foo=True)))
        self.assertTrue(Q(foo__eq=True)(Mock(foo=30)))
        self.assertTrue(Q(foo__eq=True)(Mock(foo='foo')))

        self.assertFalse(Q(foo__eq=True)(Mock(foo='')))
        self.assertFalse(Q(foo__eq=True)(Mock(foo=None)))
        self.assertFalse(Q(foo__eq=True)(Mock(foo=0)))

    def test_eq__False(self):
        self.assertFalse(Q(foo__eq=False)(Mock(foo=True)))
        self.assertFalse(Q(foo__eq=False)(Mock(foo=30)))
        self.assertFalse(Q(foo__eq=False)(Mock(foo='foo')))

        self.assertTrue(Q(foo__eq=False)(Mock(foo='')))
        self.assertTrue(Q(foo__eq=False)(Mock(foo=None)))
        self.assertTrue(Q(foo__eq=False)(Mock(foo=0)))

    def test_ne(self):
        self.assertTrue(Q(foo__ne=30)(Mock(foo=20)))
        self.assertTrue(Q(foo__ne=30)(Mock(foo=None)))

        self.assertFalse(Q(foo__ne=30)(Mock(foo=30)))

    def test_gt(self):
        self.assertTrue(Q(foo__gt=30)(Mock(foo=31)))

        self.assertFalse(Q(foo__gt=30)(Mock(foo=30)))

    def test_gte(self):
        self.assertTrue(Q(foo__gte=30)(Mock(foo=31)))
        self.assertTrue(Q(foo__gte=30)(Mock(foo=30)))

        self.assertFalse(Q(foo__gte=30)(Mock(foo=29)))

    def test_lt(self):
        self.assertTrue(Q(foo__lt=30)(Mock(foo=29)))

        self.assertFalse(Q(foo__lt=30)(Mock(foo=30)))

    def test_lte(self):
        self.assertTrue(Q(foo__lte=30)(Mock(foo=29)))
        self.assertTrue(Q(foo__lte=30)(Mock(foo=30)))

        self.assertFalse(Q(foo__lte=30)(Mock(foo=31)))

    def test_contains(self):
        self.assertTrue(
            Q(foo__contains='_')(Mock(foo='the_quick')),
        )
        self.assertFalse(
            Q(foo__contains='_')(Mock(foo='the!quick')),
        )

    def test_startswith(self):
        self.assertTrue(
            Q(foo__startswith='_')(Mock(foo='_the_quick')),
        )
        self.assertFalse(
            Q(foo__startswith='_')(Mock(foo='!the!quick')),
        )

    def test_endswith(self):
        self.assertTrue(
            Q(foo__endswith='_')(Mock(foo='the_quick_')),
        )
        self.assertFalse(
            Q(foo__endswith='_')(Mock(foo='the!quick!')),
        )

    def test_is(self):
        obj = object()
        self.assertTrue(Q(foo__is=None)(Mock(foo=None)))
        self.assertTrue(Q(foo__is=obj)(Mock(foo=obj)))
        self.assertFalse(Q(foo__is=None)(Mock(foo=0)))
        self.assertFalse(Q(foo__is=obj)(Mock(foo=object())))

    def test_is_not(self):
        obj = object()
        self.assertTrue(Q(foo__is_not=None)(Mock(foo=0)))
        self.assertTrue(Q(foo__is_not=obj)(Mock(foo=object())))
        self.assertFalse(Q(foo__is_not=None)(Mock(foo=None)))
        self.assertFalse(Q(foo__is_not=obj)(Mock(foo=obj)))

    def test_nested_fields(self):
        obj = Mock(
            author=Mock(
                full_name='George Costanza',
                is_staff=True,
                rank=3.03,
                violations=None,
            ),
            tags=['abc', 'def', 'ghi'],
        )
        self.assertTrue(
            Q(author__is_staff__eq=True,
              author__full_name__contains=' ',
              author__full_name__startswith='George',
              author__full_name__endswith='Costanza',
              author__rank__gt=2.3,
              author__violations__is=None,
              tags__contains='abc')(obj)
        )
        self.assertFalse(
            Q(author__is_staff__eq=True,
              author__full_name__contains=' ',
              author__full_name__startswith='George',
              author__full_name__endswith='Costanza',
              author__rank__gt=2.3,
              author__violations__is=None,
              tags__contains='zzz')(obj)
        )
        self.assertFalse(
            Q(author__is_staff__eq=False,
              author__full_name__contains=' ',
              author__full_name__startswith='George',
              author__full_name__endswith='Costanza',
              author__rank__gt=2.3,
              author__violations__is=None)(obj)
        )

    def test_all_must_match(self):
        obj = Mock(
            full_name='George Costanza',
            is_staff=True,
            rank=3.03,
            violations=None,
        )
        self.assertTrue(
            Q(is_staff__eq=True,
              full_name__contains=' ',
              full_name__startswith='George',
              full_name__endswith='Costanza',
              rank__gt=2.3,
              violations__is=None)(obj)
        )
        self.assertFalse(
            Q(is_staff__eq=False,
              full_name__contains=' ',
              full_name__startswith='George',
              full_name__endswith='Costanza',
              rank__gt=2.3,
              violations__is=None)(obj)
        )

    def test_nested_Q_objects(self):
        x = Mock(foo=Mock(bar=1, baz=2))
        q1 = Q(Q(foo__x__eq=True), foo__bar__eq=1) & Q(foo__baz__eq=2)
        q2 = Q(foo__bar__eq=1) & Q(foo__baz__eq=3)
        q3 = Q(foo__baz__eq=3) | Q(foo__bar__eq=1)
        q4 = Q(foo__bar__eq=1) & Q(Q(foo__x__ne=True), foo__baz__eq=2)
        self.assertTrue(q1(x))
        self.assertFalse(q2(x))
        self.assertTrue((~q2)(x))

        self.assertTrue(q3(x))
        self.assertFalse(q4(x))

    def test_now_eq__no_previous_version(self):
        class X(object):
            foo = 1
        q = Q(foo__now_eq=1)
        self.assertTrue(q(X()))

    def test_now_eq(self):
        x1 = Mock(foo=Mock(bar=1, baz=2))
        x1._previous_version = Mock(foo=Mock(bar=0, baz=2))
        q = Q(foo__bar__now_eq=1)
        self.assertTrue(q(x1))

        x2 = Mock(foo=Mock(bar=1, baz=2))
        x2._previous_version = Mock(foo=Mock(bar=1, baz=2))
        self.assertFalse(q(x2))

        x3 = Mock(state1='PUBLISHED', state2='PUBLISHED')
        x3._previous_version = Mock(
            state1='PENDING', state2='PUBLISHED',
        )
        self.assertTrue(Q(state1__now_eq='PUBLISHED')(x3))
        self.assertFalse(Q(state2__now_eq='PUBLISHED')(x3))

    def test_now_ne(self):
        x1 = Mock(foo=Mock(bar=1, baz=1))
        x1._previous_version = Mock(foo=Mock(bar=0, baz=2))
        self.assertTrue(Q(foo__bar__now_ne=2)(x1))
        self.assertFalse(Q(foo__baz__now_ne=2)(x1))

    def test_now_gt(self):
        x1 = Mock(foo=Mock(bar=22, baz=22))
        x1._previous_version = Mock(foo=Mock(bar=10, baz=42))
        self.assertTrue(Q(foo__bar__now_gt=20)(x1))
        self.assertFalse(Q(foo__baz__now_gt=20)(x1))

    def test_now_lt(self):
        x1 = Mock(foo=Mock(bar=22, baz=22))
        x1._previous_version = Mock(foo=Mock(bar=42, baz=12))
        self.assertTrue(Q(foo__bar__now_lt=40)(x1))
        self.assertFalse(Q(foo__baz__now_lt=40)(x1))

    def test_now_gte(self):
        x1 = Mock(foo=Mock(bar=22, baz=22))
        x1._previous_version = Mock(foo=Mock(bar=22, baz=42))
        self.assertTrue(Q(foo__bar__now_gte=22)(x1))
        self.assertFalse(Q(foo__baz__now_gte=22)(x1))

    def test_now_lte(self):
        x1 = Mock(foo=Mock(bar=22, baz=22))
        x1._previous_version = Mock(foo=Mock(bar=42, baz=12))
        self.assertTrue(Q(foo__bar__now_lte=22)(x1))
        self.assertFalse(Q(foo__baz__now_lte=22)(x1))

    def test_now_is(self):
        x1 = Mock(foo=Mock(bar=None, baz=None))
        x1._previous_version = Mock(foo=Mock(bar=30, baz=None))
        self.assertTrue(Q(foo__bar__now_is=None)(x1))
        self.assertFalse(Q(foo__baz__now_is=None)(x1))

    def test_now_is_not(self):
        x1 = Mock(foo=Mock(bar=30, baz=30))
        x1._previous_version = Mock(foo=Mock(bar=None, baz=10))
        self.assertTrue(Q(foo__bar__now_is_not=None)(x1))
        self.assertFalse(Q(foo__baz__now_is_not=None)(x1))

    def test_now_contains(self):
        x1 = Mock(foo=Mock(
            bar='The quick brown fox',
            baz='The quick brown fox'),
        )
        x1._previous_version = Mock(foo=Mock(
            bar='The quick red fox',
            baz='The quick brown fox',
        ))
        self.assertTrue(Q(foo__bar__now_contains='brown')(x1))
        self.assertFalse(Q(foo__baz__now_contains='brown')(x1))

    def test_now_startswith(self):
        x1 = Mock(foo=Mock(
            bar='The quick brown fox',
            baz='The quick brown fox',
        ))
        x1._previous_version = Mock(foo=Mock(
            bar='The lazy brown fox',
            baz='The quick brown fox',
        ))
        self.assertTrue(Q(foo__bar__now_startswith='The quick')(x1))
        self.assertFalse(Q(foo__baz__now_startswith='The quick')(x1))

    def test_now_endswith(self):
        x1 = Mock(foo=Mock(
            bar='The quick brown fox',
            baz='The quick brown fox',
        ))
        x1._previous_version = Mock(foo=Mock(
            bar='The lazy red fox',
            baz='The lazy brown fox',
        ))
        self.assertTrue(Q(foo__bar__now_endswith='brown fox')(x1))
        self.assertFalse(Q(foo__baz__now_endswith='brown fox')(x1))

    def test_in(self):
        x1 = Mock(foo=Mock(
            bar='PUBLISHED',
            baz='PENDING',
        ))
        self.assertTrue(Q(foo__bar__in={'PUBLISHED', 'X', 'Y', 'Z'})(x1))
        self.assertFalse(Q(foo__baz__in={'PUBLISHED', 'X', 'Y', 'Z'})(x1))

    def test_now_in(self):
        x1 = Mock(foo=Mock(
            bar='PUBLISHED',
            baz='PUBLISHED',
        ))
        x1._previous_version = Mock(foo=Mock(
            bar='PENDING',
            baz='PUBLISHED',
        ))
        self.assertTrue(Q(foo__bar__now_in={'PUBLISHED', 'X', 'Y', 'Z'})(x1))
        self.assertFalse(Q(foo__baz__now_in={'PUBLISHED', 'X', 'Y', 'Z'})(x1))

    def test_not_in(self):
        x1 = Mock(foo=Mock(
            bar='PUBLISHED',
            baz='PENDING',
        ))
        self.assertTrue(Q(foo__bar__not_in={'PENDING', 'X', 'Y', 'Z'})(x1))
        self.assertFalse(Q(foo__baz__not_in={'PENDING', 'X', 'Y', 'Z'})(x1))

    def test_now_not_in(self):
        x1 = Mock(foo=Mock(
            bar='PUBLISHED',
            baz='PUBLISHED',
        ))
        x1._previous_version = Mock(foo=Mock(
            bar='PENDING',
            baz='PUBLISHED',
        ))
        self.assertTrue(
            Q(foo__bar__now_not_in={'PENDING', 'X', 'Y', 'Z'})(x1),
        )
        self.assertFalse(
            Q(foo__baz__now_not_in={'PENDING', 'X', 'Y', 'Z'})(x1),
        )


class test_traverse_subscribers(Case):

    @patch('thorn.utils.functional.symbol_by_name')
    def test_symbol_string(self, symbol_by_name):
        symbol_by_name.return_value = 'http://e.com'
        x = [1, 2, ['!some.where.symbol', 3, 4], 5, 6]
        self.assertListEqual(list(traverse_subscribers(x)), [
            1, 2, 5, 6, symbol_by_name.return_value, 3, 4
        ])
        symbol_by_name.assert_called_once_with('some.where.symbol')

    def test_none_items(self):
        self.assertListEqual(
            list(traverse_subscribers([None, [None], None])),
            [],
        )



"""
Various data structures used in query construction.

Factored out from django.db.models.query to avoid making the main module very
large and/or so that they can be used by other modules without getting into
circular import difficulties.

"""
from __future__ import absolute_import, unicode_literals

from . import tree

__all__ = ['Q']


class Q(tree.Node):
    """Encapsulates filters as objects that can then be combined
    logically (using `&` and `|`)."""

    # Connection types
    AND = 'AND'
    OR = 'OR'
    default = AND

    def __init__(self, *args, **kwargs):
        super(Q, self).__init__(children=list(args) + list(kwargs.items()))

    def _combine(self, other, conn):
        if not isinstance(other, Q):
            raise TypeError(other)
        obj = type(self)()
        obj.connector = conn
        obj.add(self, conn)
        obj.add(other, conn)
        return obj

    def __or__(self, other):
        return self._combine(other, self.OR)

    def __and__(self, other):
        return self._combine(other, self.AND)

    def __invert__(self):
        obj = type(self)()
        obj.add(self, self.AND)
        obj.negate()
        return obj

    def clone(self):
        clone = self.__class__._new_instance(
            children=[], connector=self.connector, negated=self.negated)
        for child in self.children:
            if hasattr(child, 'clone'):
                clone.children.append(child.clone())
            else:
                clone.children.append(child)
        return clone



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-23 19:05
from __future__ import absolute_import, unicode_literals

from django.db import migrations, models
import thorn.django.models


class Migration(migrations.Migration):

    dependencies = [
        ('webhooks', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='subscriber',
            name='hmac_digest',
            field=models.CharField(
                choices=[
                    ('SHA512', 'sha512'),
                    ('SHA256', 'sha256'),
                    ('SHA1', 'sha1')],
                default='sha256',
                help_text='Specify HMAC digest type (use sha256 if uncertain)',
                max_length=64,
                verbose_name='HMAC Digest Type'),
        ),
        migrations.AddField(
            model_name='subscriber',
            name='hmac_secret',
            field=models.TextField(
                default=thorn.django.models.random_secret64,
                help_text='Specify HMAC secret for endpoints to verify',
                verbose_name='HMAC Secret'),
        ),
    ]



"""

    thorn.webhook.request
    =====================

    Webhook HTTP requests.

"""
from __future__ import absolute_import, unicode_literals

import thorn
import requests

from celery import uuid
from celery.utils import cached_property
from requests.exceptions import ConnectionError, Timeout
from requests.packages.urllib3.util.url import parse_url
from vine import maybe_promise, promise
from vine.abstract import Thenable, ThenableProxy

from ._state import app_or_default
from .utils.compat import bytes_if_py2, restore_from_keys
from .utils.log import get_logger
from .validators import deserialize_validator, serialize_validator

__all__ = ['Request']

DEFAULT_USER_AGENT = 'Mozilla/5.0 (compatible; thorn/{0}; {1})'.format(
    thorn.__version__, requests.utils.default_user_agent(),
)

REQUEST_REPR = """\
<{0}: {1.event} -> {1.subscriber.url} sender={1.sender!r}>\
"""

logger = get_logger(__name__)


@Thenable.register
class Request(ThenableProxy):
    """Webhook HTTP request

    :param event: Name of event.
    :param data: Event payload.
    :param sender: Sender of event (or :const:`None`).
    :param subscriber: Subscriber to dispatch the request for.

    :keyword on_success: Optional callback called if the HTTP request
        succeeds.  Must take single argument: ``request``.
    :keyword on_timeout: Optional callback called if the HTTP request
        times out. Must have signature: ``(request, exc)``.
    :keyword on_error: Optional callback called if the HTTP request
        fails.  Must have signature: ``(request, exc)``.
    :keyword headers: Additional HTTP headers to send with the request.
    :keyword user_agent: Set custom HTTP user agent.
    :keyword recipient_validators: List of serialized recipient validators.
    :keyword allow_keepalive: Allow reusing session for this HTTP request.
        Enabled by default.

    :keyword retry: Retry in the event of timeout/failure?
        Disabled by default.
    :keyword retry_max: Maximum number of times to retry before giving up.
        Default is 3.
    :keyword retry_delay: Delay between retries in seconds int/float.
        Default is 60 seconds.

    """

    app = None

    Session = requests.Session

    #: Holds the response after the HTTP request is performed.
    response = None

    #: Tuple of exceptions considered a connection error.
    connection_errors = (ConnectionError,)

    #: Tuple of exceptions considered a timeout error.
    timeout_errors = (Timeout,)

    #: HTTP User-Agent header.
    user_agent = DEFAULT_USER_AGENT

    def __init__(self, event, data, sender, subscriber,
                 id=None, on_success=None, on_error=None,
                 timeout=None, on_timeout=None,
                 retry=None, retry_max=None, retry_delay=None,
                 headers=None, user_agent=None, app=None,
                 recipient_validators=None,
                 allow_keepalive=True):
        self.app = app_or_default(app or self.app)
        self.id = id or uuid()
        self.event = event
        self.data = data
        self.sender = sender
        self.subscriber = subscriber
        self.timeout = timeout
        self.on_success = on_success
        self.on_timeout = maybe_promise(on_timeout)
        self.on_error = on_error
        self.retry = self.app.settings.THORN_RETRY if retry is None else retry
        self.retry_max = (
            self.app.settings.THORN_RETRY_MAX
            if retry_max is None else retry_max)
        self.retry_delay = (
            self.app.settings.THORN_RETRY_DELAY
            if retry_delay is None else retry_delay)
        if recipient_validators is None:
            recipient_validators = self.app.settings.THORN_RECIPIENT_VALIDATORS
        self.allow_keepalive = allow_keepalive
        self._recipient_validators = recipient_validators
        self.response = None
        self._headers = headers
        self._set_promise_target(promise(
            args=(self,), callback=self.on_success, on_error=self.on_error,
        ))
        if user_agent:
            self.user_agent = user_agent

    def validate_recipient(self, url):
        return [validate(url) for validate in self.recipient_validators]

    def sign_request(self, subscriber, data):
        return subscriber.sign(data)

    def dispatch(self, session=None, propagate=False, close_session=False):
        if self.cancelled:
            return
        self.validate_recipient(self.subscriber.url)
        if session is None or not self.allow_keepalive:
            session, close_session = self.Session(), True
        try:
            self.response = session.post(
                url=self.subscriber.url,
                data=self.data,
                timeout=self.timeout,
                headers=self.headers_with_hmac(
                    self.sign_request(self.subscriber, self.data),
                ),
            )
        except self.timeout_errors as exc:
            self.handle_timeout_error(exc, propagate=propagate)
        except self.connection_errors as exc:
            self.handle_connection_error(exc, propagate=propagate)
        else:
            self._p()
        finally:
            if close_session and session is not None:
                session.close()
        return self

    def handle_timeout_error(self, exc, propagate=False):
        logger.info('Timed out while dispatching webhook request: %r',
                    exc, exc_info=1, extra={'data': self.as_dict()})
        if self.on_timeout:
            return self.on_timeout(self, exc)
        return self._p.throw(exc, propagate=propagate)

    def handle_connection_error(self, exc, propagate=False):
        logger.error('Error dispatching webhook request: %r',
                     exc, exc_info=1, extra={'data': self.as_dict()})
        self._p.throw(exc, propagate=propagate)

    def as_dict(self):
        """Return a dictionary representation of this request
        suitable for serialization."""
        return {
            'id': self.id,
            'event': self.event,
            'sender': self.sender,
            'subscriber': self.subscriber.as_dict(),
            'data': self.data,
            'timeout': self.timeout,
            'retry': self.retry,
            'retry_max': self.retry_max,
            'retry_delay': self.retry_delay,
            'recipient_validators': self._serialize_validators(
                self._recipient_validators,
            ),
            'allow_keepalive': self.allow_keepalive,
        }

    def headers_with_hmac(self, hmac):
        return dict(self.headers, **{'Hook-HMAC': hmac})

    def _serialize_validators(self, validators):
        return [serialize_validator(v) for v in validators]

    def __repr__(self):
        return bytes_if_py2(REQUEST_REPR.format(type(self).__name__, self))

    def __reduce__(self):
        return restore_from_keys, (type(self), (), self.__reduce_keys__())

    def __reduce_keys__(self):
        return dict(
            self.as_dict(),
            headers=self._headers,
            user_agent=self.user_agent,
        )

    @cached_property
    def headers(self):
        return dict(self.default_headers, **self._headers or {})

    @property
    def default_headers(self):
        return {
            'Content-Type': self.subscriber.content_type,
            'User-Agent': self.user_agent,
            'Hook-Event': self.event,
            'Hook-Delivery': self.id,
        }

    @cached_property
    def urlident(self):
        """Used to order HTTP requests by URL."""
        url = parse_url(self.subscriber.url)
        return url.host, url.port or 80, url.scheme or 'http'

    @property
    def value(self):
        return self.response  # here for Thenable-compatiblity.

    @cached_property
    def recipient_validators(self):
        return [
            deserialize_validator(v) for v in self._recipient_validators
        ]



from __future__ import absolute_import, unicode_literals

import pickle

from thorn.dispatch.base import Dispatcher

from thorn.tests.case import Mock, ThornCase, call, patch


def subscriber_from_dict(d, event):
    if not isinstance(d, dict):
        d = {'url': d}
    return dict(d, event=event)


class test_Dispatcher(ThornCase):

    def setup(self):
        self._app = Mock(name='app')
        self.dispatcher = Dispatcher(app=self._app)

    def test_dispatch_request(self):
        request = Mock(name='request')
        self.dispatcher.dispatch_request(request)
        request.dispatch.assert_called_with()

    @patch('thorn.dispatch.base.barrier')
    def test_send(self, barrier):
        event = Mock(name='event')
        payload = Mock(name='payload')
        sender = Mock(name='sender')
        self.dispatcher.prepare_requests = Mock(name='prepare_requests')
        reqs = self.dispatcher.prepare_requests.return_value = [
            Mock(name='r1'), Mock(name='r2'), Mock(name='r3'),
        ]
        ret = self.dispatcher.send(event, payload, sender, timeout=3.03)
        barrier.assert_called()
        self.assertIs(ret, barrier.return_value)
        for req in reqs:
            req.dispatch.assert_called_with()
        self.assertListEqual(
            barrier.call_args[0][0], [r.dispatch() for r in reqs],
        )

    def test_prepare_requests(self):
        event = Mock(name='event')
        event.name = 'foo.bar'
        self.dispatcher.Request = Mock(name='Request')
        self.dispatcher.subscribers_for_event = Mock(name='subscribers')
        targets = self.dispatcher.subscribers_for_event.return_value = [
            Mock(name='r1'), Mock(name='r2'), Mock(name='r3'),
        ]
        targets[0].content_type = 'A'
        targets[1].content_type = 'A'
        targets[2].content_type = 'B'
        codecB = Mock(name='codecB')
        self._app.settings.THORN_CODECS = {'B': codecB}
        list(self.dispatcher.prepare_requests(
            event.name, {'foo': 'bar'}, 501, 30.3, kw=1,
        ))
        self._app.Request.assert_has_calls([
            call(event.name, {'foo': 'bar'}, 501, targets[0],
                 timeout=30.3, kw=1),
            call(event.name, {'foo': 'bar'}, 501, targets[1],
                 timeout=30.3, kw=1),
            call(event.name, codecB(), 501, targets[2],
                 timeout=30.3, kw=1),
        ])

    def test_encode_cached(self):
        cache1, cache2 = {}, {}
        encode = self.dispatcher.encode_payload = Mock(name='encode_cached')

        self.dispatcher.encode_cached('a', cache1, 'A')
        self.assertIn('A', cache1)
        v1 = cache1['A'] = object()
        self.assertEqual(self.dispatcher.encode_cached('a', cache1, 'A'), v1)
        self.assertEqual(
            self.dispatcher.encode_cached('a', cache2, 'A'), encode())

    def test_encode_payload(self):
        data = Mock(name='data')
        codec = Mock(name='codec')
        self._app.settings.THORN_CODECS = {
            'application/foo': codec,
        }
        self.assertIs(
            self.dispatcher.encode_payload(data, 'application/x-mooOOO'), data,
        )
        self.assertIs(
            self.dispatcher.encode_payload(data, 'application/foo'),
            codec.return_value,
        )
        codec.assert_called_with(data)

    def test_subscribers_for_event(self):
        source1, source2 = Mock(name='source1'), Mock(name='source2')
        source1.return_value = [1, 2, 3]
        source2.return_value = [4, 5, 6]
        self.dispatcher.subscriber_sources = [source1, source2]
        self.assertListEqual(
            list(self.dispatcher.subscribers_for_event('foo.bar')),
            [1, 2, 3, 4, 5, 6],
        )

    def test__stored_subscribers(self):
        self.assertIs(
            self.dispatcher._stored_subscribers('foo.bar'),
            self._app.Subscribers.matching.return_value,
        )
        self._app.Subscribers.matching.assert_called_with(
            event='foo.bar', user=None)

    def test_configured_subscribers__string_scalar(self):
        self._app.settings.THORN_SUBSCRIBERS = {
            'foo.bar': 'http://www.example.com/e/',
        }
        self._app.Subscriber.from_dict = subscriber_from_dict
        self.assertListEqual(
            [dict(event='foo.bar', url='http://www.example.com/e/')],
            list(self.dispatcher._configured_subscribers('foo.bar')),
        )

    def test_configured_subscribers__callback(self):
        callback1 = Mock(name='callback1')
        callback1.return_value = ['http://a.com/1', 'http://a.com/2']
        callback2 = Mock(name='callback2')
        callback2.return_value = iter(['http://b.com/1', 'http://b.com/2'])
        self._app.settings.THORN_SUBSCRIBERS = {
            'foo.bar': [callback1, callback2],
        }
        self._app.Subscriber.from_dict = subscriber_from_dict
        self.assertListEqual(
            [dict(event='foo.bar', url='http://a.com/1'),
             dict(event='foo.bar', url='http://a.com/2'),
             dict(event='foo.bar', url='http://b.com/1'),
             dict(event='foo.bar', url='http://b.com/2')],
            list(self.dispatcher._configured_subscribers('foo.bar')),
        )

    def test_configured_subscribers__string_list(self):
        self._app.settings.THORN_SUBSCRIBERS = {
            'foo.bar': [
                'http://www.example.com/e/a/',
                'http://www.example.com/e/b/',
            ],
        }
        self._app.Subscriber.from_dict = subscriber_from_dict
        self.assertListEqual(
            [dict(event='foo.bar', url='http://www.example.com/e/a/'),
             dict(event='foo.bar', url='http://www.example.com/e/b/')],
            list(self.dispatcher._configured_subscribers('foo.bar')),
        )

    def test_configured_subscribers__dict_scalar(self):
        self._app.settings.THORN_SUBSCRIBERS = {
            'foo.bar': {
                'url': 'http://www.example.com/e/',
                'content_type': 'application/x-www-form-urlencoded',
            },
        }
        self._app.Subscriber.from_dict = subscriber_from_dict
        self.assertListEqual(
            [dict(event='foo.bar',
                  url='http://www.example.com/e/',
                  content_type='application/x-www-form-urlencoded')],
            list(self.dispatcher._configured_subscribers('foo.bar')),
        )

    def test_configured_subscribers__dict_list(self):
        self._app.settings.THORN_SUBSCRIBERS = {
            'foo.bar': [
                {'url': 'http://www.example.com/e/a/',
                 'content_type': 'application/x-www-form-urlencoded'},
                {'url': 'http://www.example.com/e/b/',
                 'content_type': 'application/json'},
            ],
        }
        self._app.Subscriber.from_dict = subscriber_from_dict
        self.assertListEqual(
            [dict(event='foo.bar',
                  url='http://www.example.com/e/a/',
                  content_type='application/x-www-form-urlencoded'),
             dict(event='foo.bar',
                  url='http://www.example.com/e/b/',
                  content_type='application/json')],
            list(self.dispatcher._configured_subscribers('foo.bar')),
        )

    def test_reduce(self):
        self.dispatcher.timeout = 303
        d2 = pickle.loads(pickle.dumps(self.dispatcher))
        self.assertEqual(d2.timeout, 303)
        self.assertIs(d2.app, self.app)



"""

    thorn.funtests.celery
    =====================

    Celery application used for the Thorn Cyanide test suite.

"""
from __future__ import absolute_import, unicode_literals

import cyanide.app
import django
import os
import sys
import thorn

dist = os.path.join(
    os.path.abspath(os.path.dirname(thorn.__file__)),
    os.pardir,
)
sys.path.append(os.path.join(dist, 'testproj'))
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'testproj.settings_funtests')
django.setup()

app = cyanide.app.App(set_as_current=False)
app.cyanide_suite = 'thorn.funtests.suite:Default'
app.conf.CELERY_IMPORTS = [
    'thorn.tasks',
    'thorn.funtests.tasks',
    'cyanide.tasks',
]



from __future__ import absolute_import, unicode_literals

from thorn.environment.django import DjangoEnv

from thorn.tests.case import Case, mock, patch


class test_DjangoEnv(Case):
    Env = DjangoEnv

    def setUp(self):
        self.symbol_by_name = self.patch(
            'thorn.environment.django.symbol_by_name',
        )
        self.env = self.Env()

    @mock.environ('DJANGO_SETTINGS_MODULE', '')
    def test_autodetect__when_no_django(self):
        self.assertFalse(self.Env.autodetect())

    @mock.environ('DJANGO_SETTINGS_MODULE', 'proj.settings')
    def test_autodetect__when_django(self):
        self.assertTrue(self.Env.autodetect())

    def test_config(self):
        self.assertIs(self.env.config, self.symbol_by_name.return_value)
        self.symbol_by_name.assert_called_once_with(self.env.settings_cls)

    def test_Subscriber(self):
        self.assertIs(self.env.Subscriber, self.symbol_by_name.return_value)
        self.symbol_by_name.assert_called_once_with(self.env.subscriber_cls)

    def test_Subscribers(self):
        self.assertIs(
            self.env.Subscribers,
            self.symbol_by_name.return_value.objects,
        )
        self.symbol_by_name.assert_called_once_with(self.env.subscriber_cls)

    @patch('importlib.import_module')
    def test_signals(self, import_module):
        self.assertIs(self.env.signals, import_module.return_value)
        import_module.assert_called_with(self.env.signals_cls)

    def test_reverse(self):
        self.assertIs(self.env.reverse, self.symbol_by_name.return_value)
        self.symbol_by_name.assert_called_once_with(self.env.reverse_cls)



from __future__ import absolute_import, unicode_literals

from rest_framework.permissions import IsAuthenticated

from thorn.django.rest_framework.views import (
    SubscriberList, SubscriberDetail,
)

from thorn.tests.case import Case, Mock, SkipTest, patch


class ViewCase(Case):
    View = None
    kwargs = {}

    def setUp(self):
        if self.View is None:
            raise SkipTest('abstract test')
        self.view = self.View()
        self.view.request = Mock(name='request')
        self.view.model = Mock(name='model')
        self.view.kwargs = dict(self.kwargs)

    @patch('thorn.django.rest_framework.views.current_app')
    def test_permission_classes(self, current_app):
        app = current_app()
        app.settings.THORN_DRF_PERMISSION_CLASSES = None
        self.assertTupleEqual(self.view.permission_classes, (IsAuthenticated,))
        app.settings.THORN_DRF_PERMISSION_CLASSES = (303, 808)
        self.assertTupleEqual(self.view.permission_classes, (303, 808))


class test_SubdscriberList(ViewCase):
    View = SubscriberList

    def test_get_queryset(self):
        qs = self.view.get_queryset()
        self.view.model.objects.filter.assert_called_with(
            user=self.view.request.user,
        )
        self.assertIs(qs, self.view.model.objects.filter())

    def test_perform_create(self):
        serializer = Mock(name='serializer')
        self.view.perform_create(serializer)
        serializer.save.assert_called_with(user=self.view.request.user)


class test_SubscriberDetail(ViewCase):
    View = SubscriberDetail
    kwargs = {'uuid': 'id#1'}

    @patch('thorn.django.rest_framework.views.get_object_or_404')
    def test_get_object(self, get_object_or_404):
        subscriver = self.view.get_object()
        get_object_or_404.assert_called_with(
            self.view.model, user=self.view.request.user, uuid='id#1',
        )
        self.assertIs(subscriver, get_object_or_404())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import os

from sphinx_celery import conf

globals().update(conf.build_config(
    'thorn', __file__,
    project='Thorn',
    # version_dev='2.0',
    # version_stable='1.4',
    canonical_url='http://thorn.readthedocs.io',
    webdomain='robinhood.com',
    github_project='robinhood/thorn',
    copyright='2016',
    html_logo='images/logo.png',
    html_favicon='images/favicon.ico',
    extra_extensions=[
        'celery.contrib.sphinx',
    ],
    include_intersphinx={'python', 'sphinx', 'django', 'celery'},
    extra_intersphinx_mapping={
        'requests': ('http://requests.readthedocs.org/en/latest/', None),
    },
    html_prepend_sidebars=['sidebargithub.html'],
    django_settings='testproj.settings',
    path_additions=[os.path.join(os.pardir, 'testproj')],
    apicheck_package='thorn',
    apicheck_ignore_modules=[
        'thorn.django',
        'thorn.django.admin',
        'thorn.django.apps',
        'thorn.django.tasks',
        'thorn.django.rest_framework',
        'thorn.utils',
        'thorn.utils.django.*',
        'thorn.django.migrations.*',
        'thorn.funtests',
        'thorn.funtests.celery',
        'thorn._state',
        'thorn.generic',
    ],
))

html_theme = 'thorn'
html_theme_path = ['_theme']


def configcheck_project_settings():
    from thorn.conf import all_settings
    return all_settings()



from __future__ import absolute_import, unicode_literals

from thorn.decorators import webhook_model
from thorn.events import ModelEvent

from testapp import models

from .case import ANY, Mock, RealSignalCase, SignalCase


class test_functional_webhook_model(RealSignalCase):

    def setup(self):
        self.Model = models.Foo
        self.obj, _ = self.Model.objects.get_or_create(username='elaine')

    def teardown(self):
        self.Model.objects.all().delete()

    def test_is_dict(self):
        Model = webhook_model(on_change=ModelEvent('x.y'))(self.Model)
        self.assertIs(
            Model.webhook_events['on_change'],
            Model.webhook_events.events['on_change'],
        )

    def test_on_change(self):
        on_change = ModelEvent('x.change')
        Model = webhook_model(
            on_change=on_change,
            sender_field='username',
        )(self.Model)
        self.assertIs(Model.webhook_events.events['on_change'], on_change)
        on_change.send = Mock(name='event.send')
        self.obj.username = 'jerry'
        self.obj.save()
        on_change.send.assert_called_with(
            instance=self.obj,
            data=self.obj.webhook_payload(),
            sender=self.obj.username,
            context={'instance': self.obj.pk},
        )

    def test_on_create(self):
        on_create = ModelEvent('x.create')
        Model = webhook_model(on_create=on_create)(self.Model)
        self.assertIs(Model.webhook_events.events['on_create'], on_create)
        on_create.send = Mock(name='event.send')
        self.Model.objects.filter(username='cosmo').delete()
        obj, _ = self.Model.objects.get_or_create(username='cosmo')
        on_create.send.assert_called_with(
            instance=obj,
            data=obj.webhook_payload(),
            sender=None,
            context={'instance': obj.pk},
        )

    def test_on_delete(self):
        on_delete = ModelEvent('x.delete')
        Model = webhook_model(on_delete=on_delete)(self.Model)
        self.assertIs(Model.webhook_events.events['on_delete'], on_delete)
        on_delete.send = Mock(name='event.send')
        obj_pk = self.obj.pk
        self.obj.delete()
        on_delete.send.assert_called_with(
            instance=self.obj,
            data=self.obj.webhook_payload(),
            sender=None,
            context={'instance': obj_pk},
        )

    def test_on_custom_with_filter_dispatching_on_delete(self):
        jerry, _ = self.Model.objects.get_or_create(username='jerry')
        on_jerry_delete = ModelEvent(
            'x.delete', username__eq='jerry').dispatches_on_delete()
        Model = webhook_model(on_jerry_delete=on_jerry_delete)(self.Model)
        self.assertIs(
            Model.webhook_events.events['on_jerry_delete'],
            on_jerry_delete,
        )
        on_jerry_delete.send = Mock(name='event.send')
        self.obj.delete()
        on_jerry_delete.send.assert_not_called()

        jerry_pk = jerry.pk
        jerry.delete()

        on_jerry_delete.send.assert_called_with(
            instance=jerry,
            data=jerry.webhook_payload(),
            sender=None,
            context={'instance': jerry_pk},
        )

    def test_on_change__does_not_dispatch_on_create(self):
        on_create = ModelEvent('x.create')
        on_change = ModelEvent('x.change')
        on_create.send = Mock(name='on_create.send')
        on_change.send = Mock(name='on_change.send')

        self.Model = webhook_model(
            on_create=on_create, on_change=on_change)(self.Model)

        obj2, _ = self.Model.objects.get_or_create(username='cosmo')

        on_create.send.assert_called_with(
            instance=obj2,
            data=obj2.webhook_payload(),
            sender=None,
            context={'instance': obj2.pk},
        )
        on_change.send.assert_not_called()

    def test_contribute_to_event__reverse_propagates(self):
        reverse = Mock(name='reverse')
        reverse2 = Mock(name='reverse2')
        on_create = ModelEvent('x.create', reverse=reverse)
        on_change = ModelEvent('x.change')
        self.Model = webhook_model(
            on_create=on_create, on_change=on_change, reverse=reverse2,
        )(self.Model)
        self.assertIs(on_create.reverse, reverse)
        self.assertIs(on_change.reverse, reverse2)


class test_webhook_model(SignalCase):

    def test_with_on_create__connects(self):
        Model = webhook_model(on_create=ModelEvent('x.create'))(self.Model)
        self.post_save.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )

    def test_with_on_change__connects(self):
        Model = webhook_model(on_change=ModelEvent('x.change'))(self.Model)
        self.post_save.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )

    def test_with_on_delete__connects(self):
        Model = webhook_model(on_delete=ModelEvent('x.delete'))(self.Model)
        self.post_delete.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )

    def test_with_on_custom_delete__connects(self):
        Model = webhook_model(
            on_custom=ModelEvent('x.delete').dispatches_on_delete(),
        )(self.Model)
        self.post_delete.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )

    def test_with_on_custom_change__connects(self):
        Model = webhook_model(
            on_custom=ModelEvent('x.change').dispatches_on_change(),
        )(self.Model)
        self.post_save.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )

    def test_with_on_custom_create__connects(self):
        Model = webhook_model(
            on_custom=ModelEvent('x.create').dispatches_on_create(),
        )(self.Model)
        self.post_save.connect.assert_called_with(
            ANY, sender=Model, weak=False,
        )



"""

    thorn.dispatch.base
    ===================

    Default webhook dispatcher.

"""
from __future__ import absolute_import, unicode_literals

from functools import partial
from itertools import chain

from celery.utils.functional import maybe_list
from vine import barrier

from thorn._state import app_or_default
from thorn.generic.models import AbstractSubscriber
from thorn.utils.compat import restore_from_keys
from thorn.utils.functional import traverse_subscribers

__all__ = ['Dispatcher']


class Dispatcher(object):
    app = None

    def __init__(self, timeout=None, app=None):
        self.app = app_or_default(app or self.app)
        self.timeout = (
            timeout if timeout is not None
            else self.app.settings.THORN_EVENT_TIMEOUT
        )
        self.subscriber_sources = [
            self._configured_subscribers,
            self._stored_subscribers,
        ]

    def __reduce__(self):
        return restore_from_keys, (type(self), (), self.__reduce_keys__())

    def __reduce_keys__(self):
        return {'timeout': self.timeout}

    def send(self, event, payload, sender,
             context=None, extra_subscribers=None,
             allow_keepalive=True, **kwargs):
        return barrier([
            self.dispatch_request(req) for req in self.prepare_requests(
                event, payload, sender,
                context=context or {},
                extra_subscribers=extra_subscribers,
                allow_keepalive=allow_keepalive,
                **kwargs
            )
        ])

    def dispatch_request(self, request):
        return request.dispatch()

    def prepare_requests(self, event, payload, sender,
                         timeout=None, context=None,
                         extra_subscribers=None, **kwargs):
        # holds a cache of the payload serialized by content-type,
        # built incrementally depending on what content-types are
        # required by the subscribers.
        cache = {}
        timeout = timeout if timeout is not None else self.timeout
        return (
            self.app.Request(
                event,
                self.encode_cached(payload, cache, subscriber.content_type),
                sender, subscriber,
                timeout=timeout, **kwargs)
            for subscriber in self.subscribers_for_event(
                event, sender, context, extra_subscribers)
        )

    def encode_cached(self, payload, cache, ctype):
        try:
            return cache[ctype]
        except KeyError:
            value = cache[ctype] = self.encode_payload(payload, ctype)
            return value

    def encode_payload(self, data, content_type):
        try:
            encode = self.app.settings.THORN_CODECS[content_type]
        except KeyError:
            return data
        else:
            return encode(data)

    def subscribers_for_event(self, name,
                              sender=None, context={}, extra_subscribers=None):
        """Return a list of :class:`~thorn.django.models.Subscriber`
        subscribing to an event by name (optionally filtered by sender)."""
        return chain(*[
            source(name, sender=sender, **context)
            for source in chain(
                self.subscriber_sources,
                [partial(self._traverse_subscribers, extra_subscribers or [])],
            )
        ])

    def _maybe_subscriber(self, d, **kwargs):
        return (self.app.Subscriber.from_dict(d, **kwargs)
                if not isinstance(d, AbstractSubscriber) else d)

    def _traverse_subscribers(self, it, name, **context):
        return (self._maybe_subscriber(d, event=name)
                for d in traverse_subscribers(it, name, **context))

    def _configured_subscribers(self, name, **context):
        return self._configured_for_event(name, **context)

    def _configured_for_event(self, name, **context):
        return self._traverse_subscribers(
            maybe_list(self.app.settings.THORN_SUBSCRIBERS.get(name)) or [],
            name, **context)

    def _stored_subscribers(self, name, sender=None, **context):
        return self.app.Subscribers.matching(event=name, user=sender)



from __future__ import absolute_import, unicode_literals

from thorn import validators
from thorn.exceptions import SecurityError

from .case import Case


class test_deserialize_validator(Case):

    def test_deserialize(self):
        concrete = [
            validators.block_internal_ips(),
            validators.ensure_port(80, 443, 21, 6453),
            validators.ensure_protocol('https', 'http', 'ftp'),
            validators.block_cidr_network('192.168.0.0/16'),
        ]
        svalidators = [validators.serialize_validator(v) for v in concrete]
        re = [validators.deserialize_validator(v) for v in svalidators]

        with self.assertRaises(SecurityError):
            re[0]('127.0.0.1')
        re[0]('123.123.123.123')
        with self.assertRaises(SecurityError):
            re[1]('http://example.com:1234')
        re[1]('http://example.com')
        re[1]('http://example.com:80')
        re[1]('http://example.com:6453')
        with self.assertRaises(SecurityError):
            re[2]('gopher://mozilla')
        re[2]('http://example.com:80/path/?q=1')
        re[2]('https://example.com')
        re[2]('ftp://example.com')
        with self.assertRaises(SecurityError):
            re[3]('192.168.3.1')
        re[3]('123.123.123.123')



from __future__ import absolute_import, unicode_literals

from thorn.reverse import model_reverser

from .case import Case, Mock


class test_model_reverser(Case):

    def test_reverse(self):
        app = Mock(name='app')
        instance = Mock(name='instance')
        x = model_reverser('view-name', 'uuid', 'x.y.z', kw1='a', kw2='a.b.c')
        self.assertIs(x(instance, app=app), app.reverse.return_value)
        app.reverse.assert_called_with(
            'view-name',
            args=[instance.uuid, instance.x.y.z],
            kwargs={'kw1': instance.a, 'kw2': instance.a.b.c},
        )



from __future__ import absolute_import, unicode_literals

from django.contrib.auth import get_user_model

from thorn.django.models import Subscriber
from thorn.request import Request
from thorn.tasks import (
    send_event, dispatch_requests, dispatch_request, _worker_dispatcher,
)

from .case import (
    DEFAULT_RECIPIENT_VALIDATORS, Case, EventCase, Mock, call, patch,
)


class test_send_event(EventCase):

    @patch('thorn.tasks._worker_dispatcher')
    def test_sends_event(self, _worker_dispatcher):
        _worker_dispatcher.return_value = self.dispatcher
        send_event(self.event.name, 'foobar', 501, 3.03)
        self.dispatcher.send.assert_called_with(
            self.event.name, 'foobar', 501, timeout=3.03,
        )


class test_dispatch_requests(EventCase):

    @patch('thorn.tasks.dispatch_request')
    def test_dispatch(self, dispatch_request):
        Session = self.app.Request.Session = Mock(name='Request.Session')
        subscriber = Subscriber(url='http://example.com')
        reqs = [
            Request(self.mock_event('foo.created'), 'a', 501, subscriber,
                    timeout=3.03),
            Request(self.mock_event('foo.changed'), 'b', 501, subscriber,
                    timeout=6.06),
            Request(self.mock_event('foo.deleted'), 'c', 501, subscriber,
                    timeout=8.08),
        ]
        dispatch_requests([req.as_dict() for req in reqs])
        Session.assert_called_once_with()
        dispatch_request.assert_has_calls([
            call(session=Session(), app=self.app, **req.as_dict())
            for req in reqs
        ])


class test_dispatch_request(EventCase):

    def setUp(self):
        EventCase.setUp(self)
        self.session = Mock(name='session')
        self.user, _ = get_user_model().objects.get_or_create(
            username='test_dispatch_request',
            password='iesf&83j2aswe2',
            email='example@example.com',
        )
        self.subscriber = Subscriber(
            event='foo.created',
            url='http://example.com',
        )
        self.subscriber2 = Subscriber(
            event='foo.changed',
            url='http://example.com',
            user=self.user,
        )
        self.req = Request(
            self.subscriber.event, 'a', 501, self.subscriber, timeout=3.03)
        self.req2 = Request(
            self.subscriber2.event, 'a', 501, self.subscriber2, timeout=3.03)

    @patch('thorn.tasks.app_or_default')
    def test_success(self, app_or_default):
        _Request = app_or_default().Request
        dispatch_request(session=self.session, **self.req.as_dict())
        subscriber_dict = self.subscriber.as_dict()
        subscriber_dict.pop('user', None)
        app_or_default().Subscriber.assert_called_once_with(**subscriber_dict)
        _Request.assert_called_once_with(
            self.req.event, self.req.data,
            self.req.sender, app_or_default().Subscriber(),
            id=self.req.id, timeout=self.req.timeout, retry=self.req.retry,
            retry_max=self.req.retry_max, retry_delay=self.req.retry_delay,
            recipient_validators=DEFAULT_RECIPIENT_VALIDATORS,
            allow_keepalive=True,
        )
        _Request().dispatch.assert_called_once_with(
            session=self.session, propagate=_Request().retry)

    @patch('thorn.tasks.app_or_default')
    def test_when_keepalive_disabled(self, app_or_default):
        _Request = app_or_default().Request
        self.req.allow_keepalive = False
        dispatch_request(session=self.session, **self.req.as_dict())
        subscriber_dict = self.subscriber.as_dict()
        subscriber_dict.pop('user', None)
        app_or_default().Subscriber.assert_called_once_with(**subscriber_dict)
        _Request.assert_called_once_with(
            self.req.event, self.req.data,
            self.req.sender, app_or_default().Subscriber(),
            id=self.req.id, timeout=self.req.timeout, retry=self.req.retry,
            retry_max=self.req.retry_max, retry_delay=self.req.retry_delay,
            recipient_validators=DEFAULT_RECIPIENT_VALIDATORS,
            allow_keepalive=False,
        )
        _Request().dispatch.assert_called_once_with(
            session=self.session, propagate=_Request().retry)

    @patch('thorn.tasks.app_or_default')
    def test_success__with_user(self, app_or_default):
        _Request = app_or_default().Request
        dispatch_request(session=self.session, **self.req2.as_dict())
        subscriber_dict = self.subscriber2.as_dict()
        subscriber_dict.pop('user', None)
        app_or_default().Subscriber.assert_called_once_with(**subscriber_dict)
        _Request.assert_called_once_with(
            self.req2.event, self.req2.data,
            self.req2.sender, app_or_default().Subscriber(),
            id=self.req2.id, timeout=self.req2.timeout, retry=self.req2.retry,
            retry_max=self.req2.retry_max, retry_delay=self.req2.retry_delay,
            recipient_validators=DEFAULT_RECIPIENT_VALIDATORS,
            allow_keepalive=True,
        )
        _Request().dispatch.assert_called_once_with(
            session=self.session, propagate=_Request().retry)

    @patch('thorn.tasks.app_or_default')
    @patch('celery.app.task.Task.retry')
    def test_connection_error(self, retry, app_or_default):
        _Request = app_or_default().Request
        _Request.return_value.connection_errors = (ValueError,)
        _Request.return_value.timeout_errors = ()
        exc = _Request.return_value.dispatch.side_effect = ValueError(10)
        retry.side_effect = exc
        with self.assertRaises(ValueError):
            dispatch_request(session=self.session, **self.req.as_dict())
        retry.assert_called_with(
            exc=exc, max_retries=_Request().retry_max,
            countdown=_Request().retry_delay,
        )

    @patch('thorn.tasks.app_or_default')
    @patch('celery.app.task.Task.retry')
    def test_connection_error__retry_disabled(self, retry, app_or_default):
        _Request = app_or_default().Request
        _Request.return_value.connection_errors = (ValueError,)
        _Request.return_value.timeout_errors = ()
        _Request.return_value.retry = False
        exc = _Request.return_value.dispatch.side_effect = ValueError(11)
        retry.side_effect = exc
        self.req.retry = False
        with self.assertRaises(ValueError):
            dispatch_request(session=self.session, **self.req.as_dict())
        retry.assert_not_called()


class test_worker_dispatcher(Case):

    @patch('thorn.dispatch.celery.WorkerDispatcher')
    def test(self, WorkerDispatcher):
        _worker_dispatcher.clear()
        try:
            self.assertIs(_worker_dispatcher(), WorkerDispatcher())
        finally:
            _worker_dispatcher.clear()



from __future__ import absolute_import, unicode_literals

from thorn.dispatch.celery import Dispatcher, WorkerDispatcher

from thorn.tests.case import Case, Mock, call, patch


class test_Dispatcher(Case):

    @patch('thorn.dispatch.celery.send_event')
    def test_send(self, send_event):
        event = Mock(name='event')
        payload = Mock(name='payload')
        user = Mock(name='user')
        res = Dispatcher().send(event, payload, user, timeout=3.03, kw=9)
        send_event.s.assert_called_once_with(
            event, payload, user.pk, 3.03,
        )
        send_event.s().apply_async.assert_called_once_with()
        self.assertIs(res, send_event.s().apply_async())


class test_WorkerDispatcher(Case):

    def setUp(self):
        self.app = Mock(name='app')
        self.dispatcher = WorkerDispatcher(app=self.app)

    @patch('thorn.dispatch.celery.dispatch_requests')
    @patch('thorn.dispatch.celery.group')
    def test_send(self, group, dispatch_requests):
        def eval_genexp(x):
            list(x)
            return group.return_value
        group.side_effect = eval_genexp
        reqs = [Mock(name='r1'), Mock(name='r2'), Mock(name='r2')]
        self.dispatcher.prepare_requests = Mock(name='prepare_requests')
        self.dispatcher.prepare_requests.return_value = reqs
        self.dispatcher.group_requests = Mock(name='group_requests')
        self.dispatcher.group_requests.return_value = [
            [r] for r in reqs
        ]
        self.dispatcher.send(Mock(), Mock(), Mock(), Mock())
        dispatch_requests.s.assert_has_calls([
            call([req.as_dict()]) for req in reqs
        ])

    @patch('thorn.dispatch.celery.groupbymax')
    def test_group_requests(self, groupbymax):
        reqs = [Mock(name='r1'), Mock(name='r2'), Mock(name='r3')]
        ret = self.dispatcher.group_requests(reqs)
        groupbymax.assert_called_with(
            reqs,
            max=self.app.settings.THORN_CHUNKSIZE,
            key=self.dispatcher._compare_requests,
        )
        self.assertIs(ret, groupbymax())

    def test_compare_requests(self):
        a = Mock(name='r1')
        a.urlident = 'foo'
        b = Mock(name='r2')
        b.urlident = 'bar'
        self.assertTrue(self.dispatcher._compare_requests(a, a))
        self.assertFalse(self.dispatcher._compare_requests(a, b))



"""

    thorn.webhook.events
    ====================

    User-defined webhook events.

"""
from __future__ import absolute_import, unicode_literals

from operator import attrgetter
from six import iteritems as items, iterkeys as keys
from weakref import WeakSet

from celery.utils import cached_property

from ._state import app_or_default
from .utils.compat import bytes_if_py2, restore_from_keys
from .utils.functional import Q

__all__ = ['Event', 'ModelEvent']


def _true(*args, **kwargs):
    return True


class Event(object):
    """Webhook Event.

    :param name: Name of this event.
        Namespaces can be dot-separated, and if so subscribers can glob-match
        based on the parts in the name (e.g. ``"order.created"``).

    :keyword timeout: Default request timeout for this event.
    :keyword retry: Enable/disable retries when dispatching this event fails
        (disabled by default).
    :keyword retry_max: Max number of retries (3 by default).
    :keyword retry_delay: Delay between retries (60 seconds by default).
    :keyword recipient_validators: List of functions validating the recipient
        URL string.  Functions must return False if the URL is blocked.
        Default is to only allow HTTP and HTTPS, with respective reserved
        ports 80 and 443, and to block internal IP networks, and can
        be changed using the :setting:`THORN_RECIPIENT_VALIDATORS` setting::

            recipient_validators=[
                thorn.validators.block_internal_ips(),
                thorn.validators.ensure_protocol('http', 'https'),
                thorn.validators.ensure_port(80, 443),
            ]

        WARNING: :func:`~thorn.validators.block_internal_ips` will only
        test for reserved internal networks, and not private networks
        with a public IP address.  You can block those using
        :class:`~thorn.validators.block_cidr_network`.
    :keyword subscribers: Additional subscribers, as a list of URLs,
        subscriber model objects, or callback functions returning these
    :keyword request_data: Optional mapping of extra data to inject into
        event payloads,
    :keyword allow_keepalive: Flag to disable HTTP connection keepalive
        for this event only.  Keepalive is enabled by default.

    """
    app = None
    allow_keepalive = True
    recipient_validators = None

    def __init__(self, name,
                 timeout=None, dispatcher=None,
                 retry=None, retry_max=None, retry_delay=None, app=None,
                 recipient_validators=None, subscribers=None,
                 request_data=None, allow_keepalive=None,
                 **kwargs):
        self.name = name
        self.timeout = timeout
        self._dispatcher = dispatcher
        self.retry = retry
        self.retry_max = retry_max
        self.retry_delay = retry_delay
        self.request_data = request_data
        if allow_keepalive is not None:
            self.allow_keepalive = allow_keepalive
        if recipient_validators is not None:
            self.recipient_validators = recipient_validators
        self._subscribers = subscribers
        self.app = app_or_default(app or self.app)

    def send(self, data, sender=None,
             on_success=None, on_error=None, timeout=None, on_timeout=None):
        """Send event to all subscribers.

        :param data: Event payload (must be json serializable).

        :keyword sender: Optional event sender, as a
            :class:`~django.contrib.auth.models.User` instance.
        :keyword timeout: Specify custom HTTP request timeout
            overriding the :setting:`THORN_EVENT_TIMEOUT` setting.

        :keyword on_success: Callback called for each HTTP request
            if the request succeeds.  Must take single
            :class:`~thorn.request.Request` argument.
        :keyword on_timeout: Callback called for each HTTP request
            if the request times out.  Takes two arguments:
            a :class:`~thorn.request.Request`, and the time out
            exception instance.
        :keyword on_error: Callback called for each HTTP request
            if the request fails.  Takes two arguments:
            a :class:`~thorn.request.Request` argument, and
            the error exception instance.
        :keyword context: Extra context to pass to subscriber callbacks

        """
        return self._send(
            data,
            sender=sender, on_success=on_success, on_error=on_error,
            timeout=timeout, on_timeout=on_timeout,
        )

    def prepare_payload(self, data):
        return dict(self.request_data, **data) if self.request_data else data

    def _send(self, data, sender=None,
              on_success=None, on_error=None,
              timeout=None, on_timeout=None, context=None):
        timeout = timeout if timeout is not None else self.timeout
        return self.dispatcher.send(
            self.name, self.prepare_payload(data), sender,
            context=context,
            on_success=on_success, on_error=on_error,
            timeout=timeout, on_timeout=on_timeout, retry=self.retry,
            retry_max=self.retry_max, retry_delay=self.retry_delay,
            recipient_validators=self.prepared_recipient_validators,
            extra_subscribers=self._subscribers,
            allow_keepalive=self.allow_keepalive,
        )

    def prepare_recipient_validators(self, validators):
        return validators

    def __repr__(self):
        return bytes_if_py2("<{0}: {1} ({2:#x})>".format(
            type(self).__name__, self.name, id(self)))

    def __reduce__(self):
        return restore_from_keys, (type(self), (), self.__reduce_keys__())

    def __reduce_keys__(self):
        return {
            'name': self.name,
            'timeout': self.timeout,
            'dispatcher': self._dispatcher,
            'retry': self.retry,
            'retry_max': self.retry_max,
            'retry_delay': self.retry_delay,
            'subscribers': self._subscribers,
            'request_data': self.request_data,
            'allow_keepalive': self.allow_keepalive,
        }

    @cached_property
    def prepared_recipient_validators(self):
        return self.prepare_recipient_validators(self.recipient_validators)

    @property
    def subscribers(self):
        return self.dispatcher.subscribers_for_event(
            self.name, extra_subscribers=self._subscribers,
        )

    @subscribers.setter
    def subscribers(self, subscribers):
        self._subscribers = subscribers

    @property
    def dispatcher(self):
        return (self._dispatcher if self._dispatcher is not None
                else self.app.dispatcher)


class ModelEvent(Event):
    """Event related to model changes.

    This event type follows a specific payload format:

    .. code-block:: json

        {"event": "(str)event_name",
         "ref": "(URL)model_location",
         "sender": "(User pk)optional_sender",
         "data": {"event specific data": "value"}}

    :param name: Name of event.
    :param reverse: A function that takes a model instance and returns
        the canonical URL for that resource.
    :keyword sender_field:
        Field used as a sender for events, e.g. ``"account.user"``,
        will use ``instance.account.user``.
    :keyword $field__$op: Optional filter arguments to filter the model
        instances to dispatch for.  These keyword arguments
        can be defined just like the arguments to a Django query set,
        the only difference being that you have to specify an operator
        for every field: this means ``last_name="jerry"`` does not work,
        and you have to use ``last_name__eq="jerry"`` instead.

        See :class:`~thorn.utils.functional.Q` for more information.

    :keyword signal_dispatcher: Custom
        :class:`~thorn.django.signals.signal_dispatcher` used to
        connect this event to a model signal.

    .. seealso:

        In addition the same arguments as :class:`Event` is supported.

    """
    signal_dispatcher = None

    def __init__(self, name, *args, **kwargs):
        super(ModelEvent, self).__init__(name, **kwargs)
        self._kwargs = kwargs
        self._filterargs = args

        self.models = WeakSet()

        # initialize the filter fields: {field}__{op}
        self.filter_fields = {
            k: v for k, v in items(kwargs) if '__' in k
        }
        # optimization: Django: Transition operators require the unchanged
        # database fields before saving, a pre_save signal
        # handles this, but we want to avoid the extra database hit
        # when they are not in use.
        self.use_transitions = any(
            '__now_' in k for k in keys(self.filter_fields),
        )
        # _filterargs is set by __reduce__ to restore *args
        restored_args = kwargs.get('_filterargs') or ()
        self._init_attrs(**kwargs)
        self._filter_predicate = (
            Q(*args + restored_args, **self.filter_fields)
            if args or self.filter_fields else _true)

    def _init_attrs(self, reverse=None, sender_field=None,
                    signal_dispatcher=None, **kwargs):
        self.reverse = reverse
        self.sender_field = sender_field
        self.signal_dispatcher = signal_dispatcher

    def send(self, instance, data=None, sender=None, **kwargs):
        """Send event for model ``instance``.

        :keyword data: Event specific data.

        See :meth:`Event.send` for more arguments supported.

        """
        return self._send(self.to_message(
            data,
            sender=sender,
            ref=self.reverse(instance, app=self.app) if self.reverse else None,
        ), sender=sender, **kwargs)

    def send_from_instance(self, instance, context={}, **kwargs):
        return self.send(
            instance=instance,
            data=self.instance_data(instance),
            sender=self.instance_sender(instance),
            context=context,
        )

    def to_message(self, data, sender=None, ref=None):
        return {
            'event': self.name,
            'ref': ref,
            'sender': sender.get_username() if sender else sender,
            'data': data or {},
        }

    def instance_data(self, instance):
        """Get event data from ``instance.webhook_payload()``."""
        try:
            handler = instance.webhook_payload
        except AttributeError:
            pass
        else:
            return handler()

    def instance_sender(self, instance):
        """Get event ``sender`` from model instance."""
        if self.sender_field:
            return attrgetter(self.sender_field)(instance)

    def connect_model(self, model):
        self.models.add(model)
        self._connect_model_signal(model)

    def _connect_model_signal(self, model):
        if self.signal_dispatcher:
            self.signal_dispatcher.connect(sender=model)

    def should_dispatch(self, instance, **kwargs):
        return self._filter_predicate(instance)

    def on_signal(self, instance, **kwargs):
        if self.should_dispatch(instance, **kwargs):
            return self.send_from_instance(instance, **kwargs)

    def dispatches_on_create(self):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_create)

    def dispatches_on_change(self):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_change)

    def dispatches_on_delete(self):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_delete)

    def dispatches_on_m2m_add(self, related_field):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_m2m_add, related_field)

    def dispatches_on_m2m_remove(self, related_field):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_m2m_remove, related_field)

    def dispatches_on_m2m_clear(self, related_field):
        return self._set_default_signal_dispatcher(
            self.app.signals.dispatch_on_m2m_clear, related_field)

    def _set_default_signal_dispatcher(self, signal_dispatcher, *args):
        if self._signal_dispatcher is None:
            self._signal_dispatcher = self._prepare_signal_dispatcher(
                signal_dispatcher, *args)
        return self

    def __reduce_keys__(self):
        return dict(self._kwargs, name=self.name, _filterargs=self._filterargs)

    def _prepare_signal_dispatcher(self, signal_dispatcher, *args):
        d = signal_dispatcher(self.on_signal, *args)
        d.use_transitions = self.use_transitions
        return d

    @property
    def signal_dispatcher(self):
        return self._signal_dispatcher

    @signal_dispatcher.setter
    def signal_dispatcher(self, signal_dispatcher):
        self._signal_dispatcher = (
            self._prepare_signal_dispatcher(signal_dispatcher)
            if signal_dispatcher is not None else None
        )



from __future__ import absolute_import, unicode_literals

rtd_base = 'http://thorn.readthedocs.io/en/latest'


def rtd(*path):
    return '/'.join((rtd_base,) + path)

REFBASE = 'http://docs.celeryproject.org/en/latest'
REFS = {
    'mailing-list':
        'http://groups.google.com/group/thorn-users',
    'irc-channel': 'getting-started/resources.html#irc',
    'events-guide': rtd('userguide/events.html'),
    'reporting-bugs': 'contributing.html#reporting-bugs',
}



from __future__ import absolute_import, unicode_literals

import pickle

from thorn.conf import MIME_JSON
from thorn.exceptions import SecurityError
from thorn.request import Request

from .case import DEFAULT_RECIPIENT_VALIDATORS, EventCase, Mock, patch, skip


class test_Request(EventCase):

    def setUp(self):
        EventCase.setUp(self)
        self.req = self.mock_req(
            self.event.name, 'http://example.com:80/hook#id1?x=303',
        )
        self.gethostbyname = self.patch('socket.gethostbyname')
        self.gethostbyname.return_value = '123.123.123.123'

    def mock_req(self, event, url, **kwargs):
        kwargs.setdefault('on_success', Mock(name='on_success'))
        kwargs.setdefault('on_timeout', Mock(name='on_timeout'))
        kwargs.setdefault('on_error', Mock(name='on_error'))
        subscriber = Mock(name='subscriber')
        subscriber.url = url
        subscriber.content_type = MIME_JSON
        return Request(
            event, 'data', 'george', subscriber,
            timeout=3.03,
            **kwargs
        )

    def test_custom_user_agent(self):
        x = self.mock_req('foo.bar', 'http://e.com', user_agent='MyAgent')
        self.assertEqual(x.user_agent, 'MyAgent')
        self.assertEqual(x.headers['User-Agent'], 'MyAgent')

    def expected_headers(self, req):
        return dict(req.headers, **{
            'Hook-HMAC': req.sign_request(req.subscriber, req.data),
        })

    def test_dispatch(self):
        session = Mock(name='session')
        expected_headers = self.expected_headers(self.req)
        self.req.dispatch(session=session)
        session.post.assert_called_with(
            url=self.req.subscriber.url,
            data=self.req.data,
            headers=expected_headers,
            timeout=self.req.timeout,
        )
        self.req.on_success.assert_called_with(self.req)
        self.req.on_success = None
        self.req.dispatch(session=session)
        self.assertIs(self.req.response, session.post())
        self.assertIs(self.req.value, session.post())
        session.close.assert_not_called()

    def test_dispatch__cancelled(self):
        session = Mock(name='session')
        self.req.cancel()
        self.req.dispatch(session=session)
        session.post.assert_not_called()

    def test_dispatch__keepalive_disabled(self):
        self.req.allow_keepalive = False
        session = Mock(name='session')
        self.req.Session = Mock(name='req.Session')
        expected_headers = self.expected_headers(self.req)
        self.req.dispatch(session=session)
        session.post.assert_not_called()
        self.req.Session.assert_called_once_with()
        self.req.Session().post.assert_called_with(
            url=self.req.subscriber.url,
            data=self.req.data,
            headers=expected_headers,
            timeout=self.req.timeout,
        )
        self.req.Session().close.assert_called_once_with()

    def test_dispatch__connection_error(self):
        session = Mock(name='session')
        exc = session.post.side_effect = ValueError('foo')
        req = self.mock_req(
            self.event.name, 'http://e.com:80/hook',
            on_error=None, on_timeout=None,
        )
        req.connection_errors = (type(exc),)
        with self.assertRaises(ValueError):
            req.dispatch(session=session, propagate=True)
        req.handle_connection_error = Mock(name='handle_connection_error')
        req.dispatch(session=session, propagate=False)
        req.handle_connection_error.assert_called_with(exc, propagate=False)

    def test_dispatch__illegal_port(self):
        session = Mock(name='session')
        req = self.mock_req(
            self.event.name, 'http://e.com:1234/hook',
        )
        with self.assertRaises(SecurityError):
            req.dispatch(session=session, propagate=True)
        session.post.assert_not_called()

    def test_dispatch__timeout_error(self):
        session = Mock(name='session')
        exc = session.post.side_effect = ValueError('foo')
        req = self.mock_req(
            self.event.name, 'http://e.com:80/hook',
            on_error=None, on_timeout=None,
        )
        req.timeout_errors, req.connection_errors = (type(exc),), ()
        with self.assertRaises(ValueError):
            req.dispatch(session=session, propagate=True)

        req.handle_timeout_error = Mock(name='handle_timeout_error')
        req.dispatch(session=session, propagate=False)
        req.handle_timeout_error.assert_called_with(exc, propagate=False)

    @patch('thorn.request.logger')
    def test_handle_timeout_error(self, logger):
        exc = KeyError()
        self.req.handle_timeout_error(exc)
        self.req.on_timeout.fun.assert_called_with(self.req, exc)
        self.req.on_error.assert_not_called()

        self.req.on_timeout = None
        self.req.handle_timeout_error(exc)
        self.req.on_error.assert_called_with(self.req, exc)

        self.req.on_error = None
        self.req.handle_timeout_error(exc)
        logger.info.assert_called()

    @patch('thorn.request.logger')
    def test_handle_connection_error(self, logger):
        try:
            raise KeyError()
        except KeyError as exc:
            self.req.handle_connection_error(exc)
            self.req.on_error.assert_called_with(self.req, exc)
            logger.error.assert_called()
            self.req.on_error = None
            self.req.handle_connection_error(exc)

    def test_as_dict(self):
        self.assertDictEqual(self.req.as_dict(), {
            'id': self.req.id,
            'event': self.req.event,
            'sender': self.req.sender,
            'subscriber': self.req.subscriber.as_dict(),
            'data': self.req.data,
            'timeout': self.req.timeout,
            'retry': self.req.retry,
            'retry_delay': self.req.retry_delay,
            'retry_max': self.req.retry_max,
            'recipient_validators': DEFAULT_RECIPIENT_VALIDATORS,
            'allow_keepalive': self.req.allow_keepalive,
        })

    def test_urlident(self):
        self.assertEqual(
            self.mock_req('foo.bar', 'http://a.com/').urlident,
            self.mock_req('foo.bar', 'http://a.com:80/bar#id1?x=y').urlident,
        )
        self.assertEqual(
            self.mock_req('foo.bar', 'http://a.com/').urlident,
            self.mock_req('foo.bar', 'http://a.com').urlident,
        )
        self.assertNotEqual(
            self.mock_req('foo.bar', 'http://a.com:80/').urlident,
            self.mock_req('foo.bar', 'http://a.com:82/').urlident,
        )
        self.assertNotEqual(
            self.mock_req('foo.bar', 'https://a.com/').urlident,
            self.mock_req('foo.bar', 'http://a.com/').urlident,
        )
        self.assertNotEqual(
            self.mock_req('foo.bar', 'http://a.com/').urlident,
            self.mock_req('foo.bar', 'http://b.com/').urlident,
        )

    def test_reduce(self):

        class Subscriber(object):
            def as_dict(self):
                return {'value': 808}
        self.req.subscriber = Subscriber()
        r2 = pickle.loads(pickle.dumps(self.req))
        self.assertIs(r2.app, self.app)

    def test_repr(self):
        self.assertTrue(repr(self.req))

    @skip.if_python3()
    def test_repr__bytes_on_py2(self):
        self.assertIsInstance(repr(self.req), bytes)



from __future__ import absolute_import, unicode_literals

import threading


class _TLS(threading.local):
    current_app = None
_tls = _TLS()

default_app = None


def current_app():
    app = _tls.current_app
    if app is None:
        if default_app is None:
            from thorn.app import Thorn
            set_default_app(Thorn())
        return default_app
    return app


def set_current_app(app):
    _tls.current_app = app


def set_default_app(app):
    global default_app
    default_app = app


def app_or_default(app):
    return app if app is not None else current_app()



"""

    thorn.webhook.reverse
    =====================

    Tools for URL references.

"""
from __future__ import absolute_import, unicode_literals

from operator import attrgetter
from six import iteritems as items

from thorn._state import app_or_default

__all__ = ['model_reverser']


class model_reverser(object):
    """Describes how to get the canonical URL for a model instance.

    **Examples**

    .. code-block:: pycon

        >>> model_reverser('article-detail', uuid='uuid')
        # for an article instance will generate the URL by calling:
        >>> reverse('article_detail', kwargs={'uuid': instance.uuid})

        >>> model_reverser('article-detail', 'category.name', uuid='uuid')
        # for an article instance will generate the URL by calling:
        >>> reverse('article-detail',
        ...         args=[instance.category.name],
        ...         kwargs={'uuid': instance.uuid},
        ... )

    """

    def __init__(self, view_name, *args, **kwargs):
        self.view_name = view_name
        self.args = args
        self.kwargs = kwargs

    def __call__(self, instance, app=None, **kw):
        return app_or_default(app).reverse(
            self.view_name,
            args=[attrgetter(arg)(instance) for arg in self.args],
            kwargs={
                key: attrgetter(value)(instance)
                for key, value in items(self.kwargs)
            },
            **kw
        )



"""

    thorn.funtests.suite
    =================================

    Functional test suite.

    Instructions
    ------------

    #. Start the celery worker:

    .. code-block:: console

        $ celery -A thorn.funtests worker -l info -P eventlet -c 1000

    #. Start the development web server:

        .. code-block:: console

            $ (cd testproj; python manage.py runserver)

    #. Then execute the functional test suite:

        .. code-block:: console

            $ celery -A thorn.funtests cyanide


    For a list of tests that you can select see:

    .. code-block:: console

        $ celery -A thorn.funtests cyanide -l

"""
from __future__ import absolute_import, unicode_literals

from celery.utils.imports import qualname

from testapp.models import Article, Tag

from .base import WebhookSuite, event_url, testcase

EVENT1_REF = 'hardcoded1'
EVENT1_URL = event_url('article.changed', ref=EVENT1_REF)
EVENT2_REF = 'hardcoded2'
EVENT2_URL = event_url('article.changed', ref=EVENT2_REF)


def callback_subscribers(event, sender=None, **kwargs):
    return [
        EVENT1_URL,
        EVENT2_URL,
    ]


class Default(WebhookSuite):

    @testcase('all', iterations=1)
    def endpoints(self):
        sub1 = self.subscribe('article.created')
        sub2 = self.subscribe('article.changed')
        subs = self.list_subscriptions()
        assert sub1 in subs
        assert sub2 in subs
        self.unsubscribe(sub1['subscription'])
        subs2 = self.list_subscriptions()
        assert sub1 not in subs2
        assert sub2 in subs2
        self.unsubscribe(sub2['subscription'])
        subs3 = self.list_subscriptions()
        assert sub1 not in subs3
        assert sub2 not in subs3

    @testcase('all', iterations=1)
    def subscribe_to_article_created(self, event='article.created'):
        sub = self.subscribe('article.created')
        article = self.create_article('The quick brown fox')
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=1)
    def subscribe_to_article_changed(self, event='article.changed'):
        article = self.create_article('The quick brown fox')
        sub = self.subscribe(event)
        article = Article.objects.filter(author=self.user)[0]
        article.title = 'The lazy dog'
        article.save()
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=1)
    def subscribe_to_article_published(self, event='article.published'):
        article = self.create_article('The quick brown fox')
        sub = self.subscribe(event)
        article.state = 'PUBLISHED'
        article.save()
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=1)
    def subscribe_to_article_removed(self, event='article.removed'):
        article = self.create_article('The quick brown fox')
        sub = self.subscribe(event)
        rev = self.reverse_article(article)
        article.delete()
        self.assert_article_event_received(article, event, sub, reverse=rev)

    @testcase('all', iterations=1)
    def subscribe_to_tag_added(self, event='article.tag_added'):
        article = self.create_article('The quick brown fox')
        tag, _ = Tag.objects.get_or_create(name='kids')
        sub = self.subscribe(event)
        article.tags.add(tag)
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=1)
    def subscribe_to_tag_removed(self, event='article.tag_removed'):
        article = self.create_article('The quick brown fox')
        tag, _ = Tag.objects.get_or_create(name='kids')
        article.tags.add(tag)
        sub = self.subscribe(event)
        article.tags.remove(tag)
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=1)
    def subscribe_to_tag_all_cleared(self, event='article.tag_all_cleared'):
        article = self.create_article('The quick brown fox')
        tag1, _ = Tag.objects.get_or_create(name='kids')
        tag2, _ = Tag.objects.get_or_create(name='fun')
        article.tags.add(tag1)
        article.tags.add(tag2)
        sub = self.subscribe(event)
        article.tags.clear()
        self.assert_article_event_received(article, event, sub)

    @testcase('all', iterations=5)
    def hundred_subscribers(self, event='article.created'):
        subs = [
            self.subscribe(event, rest='&n={0}'.format(str(i)))
            for i in range(100)
        ]
        article = self.create_article('The Boring Bear')
        self.assert_article_event_received(article, event, subs[0], n=100)

    @testcase('all', iterations=1)
    def sender_mismatch_does_not_dispatch(self, event='article.changed'):
        self.token = self._login('test2', 'test2')
        article = self.create_article('The Boring Bear')
        self.subscribe(event)
        article.title = 'The Mighty Duck'
        article.save()
        self.assert_webhook_not_received()

    @testcase('all', iterations=1)
    def unsubscribe_does_not_dispatch(self, event='article.created'):
        sub = self.subscribe(event)
        article = self.create_article('Angry Bots')
        self.assert_article_event_received(article, event, sub)
        self.unsubscribe(sub['subscription'])
        self.create_article('Funky Bots')
        self.assert_webhook_not_received()

    @testcase('all', iterations=1)
    def subscriber_setting(self, event='article.changed'):
        with self.worker_subscribe_to(
                'article.changed', url=self._event_url(event)):
            article = self.create_article('Lazy Babies')
            article.title = 'Hazy Babies'
            article.save()
            self.assert_article_event_received(article, event)

    @testcase('all', iterations=1)
    def subscriber_callback_setting(self, event='article.changed'):
        with self.worker_subscribe_to(
                'article.changed', callback=qualname(callback_subscribers)):
            article = self.create_article('A brown fox')
            article.title = 'A red fox'
            article.save()
            self.assert_article_event_received(article, event, ref=EVENT1_REF)
            self.assert_article_event_received(article, event, ref=EVENT2_REF)



"""

    thorn.utils.json
    ================

    Json serialization utilities.

"""
from __future__ import absolute_import, unicode_literals

import datetime
import decimal
import uuid

from six import text_type

from celery.utils.imports import symbol_by_name

try:
    from django.utils.functional import Promise as DjangoPromise
except ImportError:  # pragma: no cover
    class DjangoPromise(object):  # noqa
        pass

__all__ = ['JsonEncoder', 'dumps']


def get_best_json(attr=None,
                  choices=['kombu.utils.json', 'simplejson', 'json']):
    for i, module in enumerate(choices):
        try:
            return symbol_by_name(
                ':'.join([module, attr]) if attr else module
            )
        except (AttributeError, ImportError):
            if i + 1 >= len(choices):
                raise
json = get_best_json()


class JsonEncoder(get_best_json('JSONEncoder')):
    # like django.core.serializers.json.JSONEncoder but preserves
    # datetime microsecond information.

    def default(self, o,
                dates=(datetime.datetime, datetime.date),
                times=(datetime.time,),
                textual=(decimal.Decimal, uuid.UUID, DjangoPromise),
                isinstance=isinstance,
                datetime=datetime.datetime,
                text_type=text_type):
        if isinstance(o, dates):
            if not isinstance(o, datetime):
                o = datetime(o.year, o.month, o.day, 0, 0, 0, 0)
            r = o.isoformat()
            if r.endswith("+00:00"):
                r = r[:-6] + "Z"
            return r
        elif isinstance(o, times):
            return o.isoformat()
        elif isinstance(o, textual):
            return text_type(o)
        else:
            return super(JsonEncoder, self).default(o)


def dumps(obj, encode=json.dumps, cls=JsonEncoder):
    return encode(obj, cls=cls)



from __future__ import absolute_import, unicode_literals

from thorn.dispatch.disabled import Dispatcher

from thorn.tests.case import Case


class test_Dispatcher(Case):

    def setUp(self):
        self.dispatcher = Dispatcher()

    def test_send(self):
        self.dispatcher.send(data={'foo': 'bar'})



"""

    thorn.generic.signals
    =====================

    Dispatching by signal.

"""
from __future__ import absolute_import, unicode_literals

from six import iteritems as items

from celery.utils.imports import symbol_by_name

__all__ = ['signal_dispatcher']


class signal_dispatcher(object):
    signals = None
    use_transitions = False

    def __init__(self, fun, **kwargs):
        self.fun = fun
        self.signals = self.load_signals(self.setup_signals())

    def setup_signals(self):
        return {}

    def load_signals(self, signals):
        return {
            symbol_by_name(sig): handler
            for sig, handler in items(signals)
        }

    def should_dispatch(self, instance, **kwargs):
        return True

    def __call__(self, instance, **kwargs):
        if self.should_dispatch(instance, **kwargs):
            return self.fun(
                instance, context=self.context(instance, **kwargs))

    def connect(self, sender=None, weak=False, **kwargs):
        [self._connect_signal(signal, handler, sender, weak, **kwargs)
         for signal, handler in items(self.signals)]

    def _connect_signal(self, signal, handler, sender, weak, **kwargs):
        signal.connect(
            handler,
            sender=self.prepare_sender(sender),
            weak=weak,
            **kwargs)

    def context(self, instance, **kwargs):
        return {'instance': instance.pk}

    def prepare_sender(self, sender):
        return sender



from __future__ import absolute_import, unicode_literals

import pickle

from django.db.models.query import Q as _Q_

from thorn.django import signals
from thorn.reverse import model_reverser
from thorn.events import Event, ModelEvent, _true
from thorn.utils.functional import Q

from .case import EventCase, Mock


class test_Event(EventCase):

    def test_init(self):
        self.assertEqual(self.event.name, 'george.costanza')

    def test_subscribers(self):
        subscribers = self.event.subscribers
        self.dispatcher.subscribers_for_event.assert_called_with(
            self.event.name, extra_subscribers=None,
        )
        self.assertIs(subscribers, self.dispatcher.subscribers_for_event())

    def test_init__recipient_validators(self):
        x = Event('x.y', recipient_validators=[1])
        self.assertListEqual(x.recipient_validators, [1])

    def test_subscribers__setter(self):
        subs = [1, 2, 3]
        self.event.subscribers = subs
        self.assertIs(self.event._subscribers, subs)

    def test_repr(self):
        self.assertTrue(repr(self.event))

    def test_send(self):
        on_success = Mock(name='on_success')
        on_timeout = Mock(name='on_timeout')
        on_error = Mock(name='on_error')
        sender = Mock(name='sender')
        self.event.send(
            {'foo': 'bar'}, sender=sender, timeout=3.34,
            on_success=on_success, on_timeout=on_timeout, on_error=on_error,
        )
        self.dispatcher.send.assert_called_with(
            self.event.name, {'foo': 'bar'}, sender,
            on_success=on_success, on_error=on_error,
            timeout=3.34, on_timeout=on_timeout,
            retry=None, retry_delay=None, retry_max=None,
            recipient_validators=None,
            context=None, extra_subscribers=None, allow_keepalive=True,
        )

    def test_send__with_request_data(self):
        event = self.mock_event('x.y', request_data={'agent': 'AGENT'})
        event.send({'foo': 'bar'})
        self.dispatcher.send.assert_called_with(
            event.name, {'foo': 'bar', 'agent': 'AGENT'}, None,
            on_success=None, on_error=None,
            timeout=None, on_timeout=None,
            retry=None, retry_delay=None, retry_max=None,
            recipient_validators=None,
            context=None, extra_subscribers=None, allow_keepalive=True,
        )

    def test_send__with_disable_keepalive(self):
        event = self.mock_event('x.y', allow_keepalive=False)
        event.send({'foo': 'bar'})
        self.dispatcher.send.assert_called_with(
            event.name, {'foo': 'bar'}, None,
            on_success=None, on_error=None,
            timeout=None, on_timeout=None,
            retry=None, retry_delay=None, retry_max=None,
            recipient_validators=None,
            context=None, extra_subscribers=None, allow_keepalive=False,
        )

    def test_dispatcher(self):
        event = Event('george.costanza', app=Mock(name='app'))
        self.assertIs(event.dispatcher, event.app.dispatcher)

    def test_reduce(self):
        self.event._dispatcher = None
        e2 = pickle.loads(pickle.dumps(self.event))
        self.assertEqual(e2.name, self.event.name)
        self.assertIs(e2.app, self.app)


class test_ModelEvent(EventCase):

    def test_send(self):
        sender = Mock(name='sender')
        self.event._send = Mock(name='_send')
        self.event.app = Mock(name='app')
        instance = Mock(name='instance')
        self.event.send(instance, {'foo': 'bar'}, sender=sender)
        self.event.app.reverse.assert_called_with(
            self.event.reverse.view_name,
            args=[], kwargs={'uuid': instance.uuid},
        )
        self.event._send.assert_called_with(
            {
                'event': self.event.name,
                'sender': sender.get_username(),
                'ref': self.event.app.reverse.return_value,
                'data': {'foo': 'bar'},
            },
            sender=sender,
        )

    def test_send__with_request_data(self):
        event = self.mock_event('x.y', request_data={'agent': 'AGENT'})
        event.reverse = None
        instance = Mock(name='instance')
        event.send(instance, {'foo': 'bar'}, sender=None)
        self.dispatcher.send.assert_called_with(
            event.name,
            {
                'ref': None,
                'data': {
                    'foo': 'bar',
                },
                'agent': 'AGENT',
                'event': 'x.y',
                'sender': None,
            },
            None,
            on_success=None, on_error=None,
            timeout=None, on_timeout=None,
            retry=None, retry_delay=None, retry_max=None,
            recipient_validators=None,
            context=None, extra_subscribers=None, allow_keepalive=True,
        )

    def test_dispatches_on_create(self):
        event = self.mock_event('x.on_create').dispatches_on_create()
        self.assertIsInstance(
            event.signal_dispatcher, signals.dispatch_on_create,
        )
        self.assertEqual(event.signal_dispatcher.fun, event.on_signal)

    def test_dispatches_on_change(self):
        event = self.mock_event('x.on_change')
        event.on_signal = Mock()
        event.dispatches_on_change()
        self.assertIsInstance(
            event.signal_dispatcher, signals.dispatch_on_change,
        )
        event.signal_dispatcher.fun()
        event.on_signal.assert_called_with()
        self.assertEqual(event.signal_dispatcher.fun, event.on_signal)

    def test_dispatches_on_delete(self):
        event = self.mock_event('x.on_change').dispatches_on_delete()
        self.assertIsInstance(
            event.signal_dispatcher, signals.dispatch_on_delete,
        )
        self.assertEqual(event.signal_dispatcher.fun, event.on_signal)

    def mock_event(self, name, *args, **kwargs):
        event = ModelEvent(
            name,
            dispatcher=kwargs.get('dispatcher') or self.dispatcher,
            reverse=model_reverser('something-view', uuid='uuid'),
            *args, **kwargs
        )
        event.reverse._reverse = Mock(name='event.reverse._reverse')
        return event

    def test_default_filter_predicate(self):
        event = self.mock_event('foo.created')
        self.assertIs(event._filter_predicate, _true)

    def test_filter_predicate__filter_args(self):
        filter_fields = {'fieldA__eq': 30, 'fieldB__eq': 60}
        event = self.mock_event('foo.created', **filter_fields)
        self.assertIsInstance(event._filter_predicate, Q)

    def test_supports_django_Q_objects(self):
        m = Mock()
        m.account.user.username = 'George'
        self.event = self.mock_event('x.y', (
            _Q_(account__user__username__eq='George') |
            _Q_(account__user__username__eq='Elaine')
        ))
        self.assertTrue(self.event.should_dispatch(m))

        m.account.user.username = 'Jerry'
        self.assertFalse(self.event.should_dispatch(m))

        m.account.user.username = 'Elaine'
        self.assertTrue(self.event.should_dispatch(m))

    def test_custom_signal_dispatcher(self):
        dispatcher = Mock(name='dispatcher')
        event = self.mock_event('foo.created', signal_dispatcher=dispatcher)
        dispatcher.assert_called_with(event.on_signal)
        self.assertIs(event.signal_dispatcher, dispatcher())

        event.dispatches_on_create()
        event.dispatches_on_change()
        event.dispatches_on_delete()

        self.assertIs(event.signal_dispatcher, dispatcher())

        event.connect_model(self.Model)
        event.signal_dispatcher.connect.assert_called_with(sender=self.Model)

    def test_connect_model__no_dispatcher(self):
        event = self.mock_event('foo.custom')
        event.connect_model(self.Model)

    def test_instance_data__undefined(self):
        self.assertFalse(self.event.instance_data(object()))

    def test_instance_data__defined(self):
        instance = self.Model()
        self.assertIs(
            self.event.instance_data(instance),
            instance.webhook_payload.return_value,
        )

    def test_instance_sender__field_undefined(self):
        self.assertFalse(self.event.instance_sender(self.Model()))

    def test_instance_sender__field_defined(self):
        event = self.mock_event('x.y', sender_field='account.user')
        instance = self.Model()
        self.assertIs(
            event.instance_sender(instance),
            instance.account.user,
        )

    def test_on_signal(self):
        instance = self.Model()
        event = self.mock_event('x.y', sender_field='x.y.z.account')
        event.dispatches_on_change()
        event.send = Mock(name='ModelEvent.send')
        event.on_signal(instance)
        event.send.assert_called_with(
            instance=instance,
            data=instance.webhook_payload.return_value,
            sender=instance.x.y.z.account,
            context={},
        )

    def test_on_signal__no_sender_field(self):
        instance = self.Model()
        event = self.mock_event('x.y', sender_field=None)
        event.dispatches_on_change()
        event.send = Mock(name='ModelEvent.send')
        event.on_signal(instance)
        event.send.assert_called_with(
            instance=instance,
            data=instance.webhook_payload.return_value,
            sender=None,
            context={},
        )

    def test_reduce(self):
        self.event._kwargs['dispatcher'] = None
        self.event.reverse._reverse = None
        e2 = pickle.loads(pickle.dumps(self.event))
        self.assertEqual(e2.name, self.event.name)
        self.assertIs(e2.app, self.app)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-27 19:33
from __future__ import absolute_import, unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import thorn.generic.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Subscriber',
            fields=[
                ('id', models.AutoField(
                    auto_created=True, primary_key=True,
                    serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(
                    default=uuid.uuid4, editable=False,
                    help_text='Unique identifier for this subscriber.',
                    unique=True, verbose_name='UUID')),
                ('event', models.CharField(
                    db_index=True, help_text='Name of event to connect with',
                    max_length=190, verbose_name='event')),
                ('url', models.URLField(
                    db_index=True, help_text='Callback URL',
                    max_length=190, verbose_name='URL')),
                ('content_type', models.CharField(
                    choices=[('application/json', 'application/json'),
                             ('application/x-www-form-urlencoded',
                              'application/x-www-form-urlencoded')],
                    default='application/json',
                    help_text='''
                        Desired content type for requests to this callback.
                    '''.strip(),
                    max_length=190, verbose_name='content type')),
                ('created_at', models.DateTimeField(
                    auto_now_add=True, verbose_name='created at')),
                ('updated_at', models.DateTimeField(
                    auto_now=True, verbose_name='updated_at')),
                ('user', models.ForeignKey(
                    null=True, on_delete=django.db.models.deletion.CASCADE,
                    related_name='webhooks_subscriber',
                    to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['url', '-created_at'],
                'get_latest_by': 'updated_at',
                'verbose_name': 'subscriber',
                'verbose_name_plural': 'subscriber',
            },
            bases=(models.Model, thorn.generic.models.SubscriberModelMixin),
        ),
        migrations.AlterUniqueTogether(
            name='subscriber',
            unique_together=set([('url', 'event')]),
        ),
    ]



from __future__ import absolute_import, unicode_literals

import pickle
import thorn.dispatch.disabled
import thorn.dispatch.base

from thorn import Thorn
from thorn.decorators import webhook_model
from thorn.events import Event, ModelEvent
from thorn.reverse import model_reverser
from thorn.request import Request

from .case import Mock, ThornCase, patch


class Subscriber(object):

    def as_dict(self):
        return {}


class test_Thorn(ThornCase):

    @patch('thorn._state.set_default_app')
    def test_set_default(self, set_default_app):
        self.app.set_default()
        set_default_app.assert_called_once_with(self.app)

    def test_dispatcher__argument(self):

        class Dispatcher(object):
            app = None

        app = Thorn(set_as_current=False, dispatcher=Dispatcher)
        self.assertIsInstance(app.dispatcher, Dispatcher)
        self.assertIs(app.dispatcher.app, app)

    def test_dispatcher__setting(self):
        self.app.settings.THORN_DISPATCHER = 'disabled'
        self.assertIsInstance(
            self.app.dispatcher, thorn.dispatch.disabled.Dispatcher,
        )
        self.assertIs(self.app.dispatcher.app, self.app)

    def test_dispatcher__default(self):
        self.assertIsInstance(
            self.app.dispatcher, thorn.dispatch.base.Dispatcher,
        )

    def test_dispatcher__pickle(self):
        self.assertIs(
            pickle.loads(pickle.dumps(self.app.dispatcher)).app,
            self.app,
        )

    def test_Subscriber(self):
        self.app.env = Mock(name='env')
        self.assertIs(self.app.Subscriber, self.app.env.Subscriber)

    def test_Subscribers(self):
        self.app.env = Mock(name='env')
        self.assertIs(self.app.Subscribers, self.app.env.Subscribers)

    def test_signals(self):
        self.app.env = Mock(name='env')
        self.assertIs(self.app.signals, self.app.env.signals)

    def test_reverse(self):
        self.app.env = Mock(name='env')
        self.assertIs(self.app.reverse, self.app.env.reverse)

    def test_Event(self):
        event = self.app.Event('article.created')
        self.assertIsInstance(event, self.app.Event)
        self.assertIsInstance(event, Event)
        self.assertIs(event.app, self.app)
        self.assertIs(self.app.Event.app, self.app)

    def test_Event__pickle(self):
        event = self.app.Event('article.created')
        self.assertIs(pickle.loads(pickle.dumps(event)).app, self.app)

    def test_ModelEvent(self):
        event = self.app.ModelEvent('article.created')
        self.assertIsInstance(event, self.app.ModelEvent)
        self.assertIsInstance(event, ModelEvent)
        self.assertIs(event.app, self.app)
        self.assertIs(self.app.ModelEvent.app, self.app)

    def test_ModelEvent__pickle(self):
        event = self.app.ModelEvent('article.created')
        self.assertIs(pickle.loads(pickle.dumps(event)).app, self.app)

    def test_Request(self):
        request = self.app.Request('x.y', {}, None, Mock(name='subscriber'))
        self.assertIsInstance(request, self.app.Request)
        self.assertIsInstance(request, Request)
        self.assertIs(request.app, self.app)
        self.assertIs(self.app.Request.app, self.app)

    def test_Request__pickle(self):
        request = self.app.Request('x.y', {}, None, Subscriber())
        request2 = pickle.loads(pickle.dumps(request))
        self.assertIs(request2.app, self.app)

    def test_webhook_model(self):
        self.assertIs(self.app.webhook_model, webhook_model)

    def test_model_reverser(self):
        self.assertIs(self.app.model_reverser, model_reverser)

    def test_subclass_with_self__keep_reduce(self):
        class Object(object):
            app = None

            def __reduce__(self):
                return 303

        AppObject = self.app.subclass_with_self(Object, keep_reduce=True)
        self.assertEqual(AppObject().__reduce__(), 303)



from __future__ import absolute_import, unicode_literals

import socket

from ipaddress import ip_address, ip_network
from six import text_type

from kombu.utils.url import urlparse

from .exceptions import SecurityError

__all__ = [
    'ensure_protocol', 'ensure_port',
    'block_internal_ips', 'block_cidr_network',
]

validators = {}


def validator(fun):
    """Make validator json serializable."""
    validators[fun.__name__] = fun
    return fun


def serialize_validator(v):
    return (v._validator, v._args)


def deserialize_validator(v):
    if isinstance(v, (list, tuple)):
        name, args = v
        return validators[name](*args)
    return v


def _is_internal_address(addr):
    return any([
        addr.is_private,
        addr.is_reserved,
        addr.is_loopback,
        addr.is_multicast,
    ])


@validator
def ensure_protocol(*allowed):
    """Only allow recipient URLs using specific protocols.

    Example:
        ensure_protocol('https', 'http://')

    """
    allowed = tuple(
        x if '://' in x else x + '://'
        for x in allowed
    )

    def validate_protocol(recipient_url):
        if not recipient_url.startswith(allowed):
            raise SecurityError(
                'Protocol of recipient URL not allowed ({0} only)'.format(
                    allowed))
    validate_protocol._args = allowed
    validate_protocol._validator = 'ensure_protocol'
    return validate_protocol


@validator
def ensure_port(*allowed):
    allowed = tuple(int(p) for p in allowed)

    def validate_port(recipient_url):
        port = urlparse(recipient_url).port
        if port and int(port) not in allowed:
            raise SecurityError(
                'Port of recipient URL {0} not allowed ({1} only)'.format(
                    recipient_url, allowed))
    validate_port._args = allowed
    validate_port._validator = 'ensure_port'
    return validate_port


def _url_ip_address(url):
    try:
        return ip_address(text_type(url))
    except ValueError:
        host = urlparse(url).hostname
        return ip_address(text_type(socket.gethostbyname(host)))


@validator
def block_internal_ips():
    """Block recipient URLs that have an internal IP address.

    .. warning::

        This does not check for *private* networks, it will only
        make sure the IP address is not in a reserved private block
        (e.g. 192.168.0.1/24).

    """

    def validate_not_internal_ip(recipient_url):
        addr = _url_ip_address(recipient_url)
        if _is_internal_address(addr):
            raise SecurityError(
                'IP address of recipient {0}={1} considered private!'.format(
                    recipient_url, addr))
    validate_not_internal_ip._args = ()
    validate_not_internal_ip._validator = 'block_internal_ips'
    return validate_not_internal_ip


@validator
def block_cidr_network(*blocked_networks):
    """Block recipient URLs from a list of CIDR networks.

    Example:

        block_cidr_network('192.168.0.0/24', '132.34.23.0/24')

    """
    _blocked_networks = [ip_network(text_type(x)) for x in blocked_networks]

    def validate_cidr(recipient_url):
        recipient_addr = _url_ip_address(recipient_url)
        for blocked_network in _blocked_networks:
            if recipient_addr in blocked_network:
                raise SecurityError(
                    'IP address of recipient {0}={1} is in network {2}'.format(
                        recipient_url, recipient_addr, blocked_network,
                    ))
    validate_cidr._args = blocked_networks
    validate_cidr._validator = 'block_cidr_network'
    return validate_cidr



"""

    thorn.dispatch.disabled
    =======================

    Dispatcher doing nothing.

"""
from __future__ import absolute_import, unicode_literals

from . import base

__all__ = ['Dispatcher']


class Dispatcher(base.Dispatcher):

    def send(self, *args, **kwargs):
        pass



"""

    thorn.webhook.decorators
    ========================

    Webhook decorators.

"""
from __future__ import absolute_import, unicode_literals

from itertools import chain

from six import iteritems as items, itervalues as values
from six.moves import UserDict


class webhook_model(UserDict):
    """Decorates models to send webhooks based changes to that model.

    :keyword on_create: Event to dispatch whenever an instance of
        this model is created (``post_save``).
    :keyword on_change: Event to dispatch whenever an instance of
        this model is changed (``post_save``).
    :keyword on_delete: Event to dispatch whenever an instance of
        this model is deleted (``post_delete``).
    :keyword on_$event: Additional user defined events.,
    :keyword sender_field:
        Default field used as a sender for events, e.g. ``"account.user"``,
        will use ``instance.account.user``.

        Individual events can override the sender field user.
    :keyword reverse:
        A :class:`thorn.reverse.model_reverser` instance (or any callable
        taking an model instance as argument), that describes how to get
        the URL for an instance of this model.

        Individual events can override the reverser used.

    **Examples**

    Simple article model, where the URL reference is retrieved
    by ``reverse('article-detail', kwargs={'uuid': article.uuid})``:

    .. code-block:: python

        @webhook_model(
            on_create=ModelEvent('article.created'),
            on_change=ModelEvent('article.changed'),
            on_delete=ModelEvent('article.removed'),
            on_deactivate=ModelEvent(
                'article.deactivate', deactivated__eq=True,
            )
            reverse=model_reverser('article-detail', uuid='uuid'),
        )
        class Article(models.Model):
            uuid = models.UUIDField()

    The URL may not actually exist after deletion, so maybe we want
    to point the reference to something else in that special case,
    like a category that can be reversed by doing
    ``reverse('category-detail', args=[article.category.name])``.

    We can do that by having the ``on_delete`` event override
    the reverser used for that event only:

    .. code-block:: python

        @webhook_model(
            on_create=ModelEvent('article.created'),
            on_change=ModelEvent('article.changed'),
            on_delete=ModelEvent(
                'article.removed',
                reverse=model_reverser('category-detail', 'category.name'),
            ),

            on_hipri_delete=ModelEvent(
                'article.internal_delete', priority__gte=30.0,
            ).dispatches_on_delete(),

            reverse=model_reverser('article-detail', uuid='uuid'),
        )
        class Article(model.Model):
            uuid = models.UUIDField()
            category = models.ForeignKey('category')

    """

    def __init__(self,
                 on_create=None, on_change=None, on_delete=None,
                 reverse=None, sender_field=None, **kwargs):
        self.reverse = reverse
        self.sender_field = sender_field
        self.events = {}

        self.update_events(
            {k: v for k, v in items(kwargs) if k.startswith('on_')},
            on_create=on_create and on_create.dispatches_on_create(),
            on_change=on_change and on_change.dispatches_on_change(),
            on_delete=on_delete and on_delete.dispatches_on_delete(),
        )

    def update_events(self, events, **kwargs):
        self.events.update(self.connect_events(events, **kwargs))

    def connect_events(self, events, **kwargs):
        return {
            k: self.contribute_to_event(v)
            for k, v in chain(items(events), items(kwargs))
        }

    def contribute_to_event(self, event):
        if event:
            if event.reverse is None:
                event.reverse = self.reverse
            event.sender_field = self.sender_field
        return event

    def __call__(self, model):
        [event.connect_model(model) for event in values(self.events) if event]
        model.webhook_events = self
        return model

    @property
    def data(self):
        # for UserDict
        return self.events



from __future__ import absolute_import, unicode_literals

from operator import attrgetter, methodcaller

from celery.utils import cached_property
from celery.utils.imports import symbol_by_name
from celery.utils.functional import first

from . import _state
from .utils.compat import bytes_if_py2


def _unpickle_appattr(reverse_name, kwargs):
    return attrgetter(reverse_name)(_state.current_app())(**kwargs)


class Thorn(object):
    event_cls = 'thorn.events:Event'
    model_event_cls = 'thorn.events:ModelEvent'
    settings_cls = 'thorn.conf:Settings'
    request_cls = 'thorn.request:Request'

    dispatchers = {
        'default': 'thorn.dispatch.base:Dispatcher',
        'celery': 'thorn.dispatch.celery:Dispatcher',
        'disabled': 'thorn.dispatch.disabled:Dispatcher',
    }
    environments = {
        'thorn.environment.django:DjangoEnv',
    }

    def __init__(self, dispatcher=None, set_as_current=True):
        self._dispatcher = dispatcher
        if set_as_current:
            self.set_current()

    def set_current(self):
        _state.set_current_app(self)

    def set_default(self):
        _state.set_default_app(self)

    def autodetect_env(self, apply=methodcaller('autodetect')):
        return first(apply, map(symbol_by_name, self.environments))()

    def _get_dispatcher(self, dispatcher=None):
        if dispatcher is None:
            dispatcher = self.settings.THORN_DISPATCHER
        return symbol_by_name(dispatcher, self.dispatchers)

    @cached_property
    def dispatcher(self):
        return self.Dispatcher()

    @cached_property
    def Dispatcher(self):
        return self.subclass_with_self(self._get_dispatcher(self._dispatcher))

    @property
    def config(self):
        return self.env.config

    @property
    def Subscriber(self):
        return self.env.Subscriber

    @property
    def Subscribers(self):
        return self.env.Subscribers

    @property
    def signals(self):
        return self.env.signals

    @property
    def reverse(self):
        return self.env.reverse

    @cached_property
    def Settings(self):
        return self.subclass_with_self(self.settings_cls)

    @cached_property
    def settings(self):
        return self.Settings()

    @cached_property
    def env(self):
        return self.autodetect_env()

    @cached_property
    def Event(self):
        return self.subclass_with_self(self.event_cls)

    @cached_property
    def ModelEvent(self):
        return self.subclass_with_self(self.model_event_cls)

    @cached_property
    def webhook_model(self):
        return symbol_by_name('thorn.decorators.webhook_model')

    @cached_property
    def model_reverser(self):
        return symbol_by_name('thorn.reverse.model_reverser')

    @cached_property
    def Request(self):
        return self.subclass_with_self(self.request_cls)

    def subclass_with_self(self, Class,
                           name=None, attribute='app',
                           reverse=None, keep_reduce=False, **kw):
        """Subclass an app-compatible class by setting its app attribute
        to this instance.

        App-compatible means the class has an 'app' attribute providing
        the default app, e.g.: ``class Foo(object): app = None``.

        :param Class: The class to subclass.
        :keyword name: Custom name for the target subclass.
        :keyword attribute: Name of the attribute holding the app.
            Default is ``"app"``.
        :keyword reverse: Reverse path to this object used for pickling
            purposes.  E.g. for ``app.AsyncResult`` use ``"AsyncResult"``.
        :keyword keep_reduce: If enabled a custom ``__reduce__`` implementation
           will not be provided.

        """
        Class = symbol_by_name(Class)
        reverse = reverse if reverse else Class.__name__

        def __reduce__(self):
            return _unpickle_appattr, (reverse, self.__reduce_keys__())

        attrs = dict({attribute: self},
                     __module__=Class.__module__,
                     __doc__=Class.__doc__,
                     **kw)
        if not keep_reduce:
            attrs['__reduce__'] = __reduce__

        return type(bytes_if_py2(name or Class.__name__), (Class,), attrs)



from __future__ import absolute_import, unicode_literals

from thorn.utils import log

from thorn.tests.case import Case, patch


class test_get_logger(Case):

    @patch('thorn.utils.log._get_logger')
    def test_sets_parent(self, _get_logger):
        x = log.get_logger(__name__)
        _get_logger.assert_called_with(__name__)
        self.assertIs(x, _get_logger())
        self.assertIs(x.parent, log.base_logger)



"""

    thorn.utils.functional
    ======================

    Functional-style utilities.

"""
from __future__ import absolute_import, unicode_literals

import operator

from collections import Callable, deque
from functools import partial
from six import string_types

from celery.utils import cached_property
from celery.utils.functional import is_list, maybe_list
from celery.utils.imports import symbol_by_name

try:
    from django.db.models.query import Q as _Q_
except ImportError:  # pragma: no cover
    from .django.query_utils import Q as _Q_  # noqa

__all__ = ['groupbymax', 'Q']

E_FILTER_FIELD_MISSING_OP = """\
filter field argument {0!r} not allowed: did you mean '{0}__eq'?\
"""


def not_contains(a, b):
    """``not_contains(a, b) -> b not in a``"""
    return b not in a


def startswith(a, b):
    """``startswith(a, b) -> a.startswith(b)``"""
    return a.startswith(b)


def endswith(a, b):
    """``endswith(a, b) -> a.endswith(b)``"""
    return a.endswith(b)


def reverse_n(N, tup):
    """Reverse n first elements in a tuple."""
    return tuple(reversed(tup[:N])) + tup[N:] if N else tuple(reversed(tup))


def negate(fun):
    """Return function negating the value of a boolean function."""

    def negated(*args, **kwargs):
        return not fun(*args, **kwargs)
    return negated


def reverse_arguments(N):
    """Returns transformed function where the first N arguments are
    reversed."""

    def _inner(fun):
        def reversed(*args, **kwargs):
            return fun(*reverse_n(N, args), **kwargs)
        return reversed
    return _inner


def traverse_subscribers(it, *args, **kwargs):
    stream = deque([it])
    while stream:
        for node in maybe_list(stream.popleft()):
            if isinstance(node, string_types) and node.startswith('!'):
                node = symbol_by_name(node[1:])
            if isinstance(node, Callable):
                node = node(*args, **kwargs)
            if is_list(node):
                stream.append(node)
            elif node:
                yield node


def wrap_transition(op, did_change):
    """Transforms operator into a transition operator, i.e. one that
    only returns true if the ``did_change`` operator also returns true.

    E.g. ``wrap_transition(operator.eq, operator.ne)`` returns function
    with signature ``(new_value, needle, old_value)`` and only returns
    true if new_value is equal to needle, but old_value was not equal
    to needle.

    """

    def compare(new_value, needle, old_value):
        return did_change(old_value, needle) and op(new_value, needle)

    return compare


def groupbymax(it, max, key=operator.eq, sentinel=object()):
    """Given an iterator emitting items in sorted order, this will
    group items together based on the key function, and produces
    one list for each group.

    :param it: Iterator emitting item in order.
    :param max: Maximum size of any group (mandatory).
    :keyword key: Function used to compare items.
        Defaults to :func:`operator.eq` matching values exactly.

    Examples:

    .. code-block:: pycon

        >>> x = ['A', 'A', 'A', 'B', 'C', 'D', 'D', 'E']
        >>> list(groupbymax(x, 3))
        [['A', 'A', 'A'], ['A'], ['B'], ['C'], ['D', 'D'], ['E']]

        # NOTE: Not technically sorted, but similar items appear in the
        # order we're matching for.
        >>> x = [('foo:A', 'foo:B', 'bar:C', 'baz:D', 'baz:E', 'baz:F']
        >>> list(groupbymax(x, 10,
        ...     key=lambda a, b: a.split(':')[0] == b.split(':')[0]))
        [['foo:A', 'foo:B'], ['bar:C'], ['baz:D', 'baz:E', 'baz:F']]

    """
    it = iter(it)
    for item in it:
        buf = []
        while 1:
            nxt = next(it, sentinel)
            if nxt is sentinel or (
                    not key(nxt, item) or len(buf) >= max - 1):
                yield [item] + buf if buf else [item]
                if nxt is not sentinel:
                    yield [nxt]
                break
            buf.append(nxt)


class Q(_Q_):
    """Object query node.

    This class works like :class:`django.db.models.Q`, but is used for
    filtering regular Python objects instead of database rows.

    **Examples**

    - Match object with ``last_name`` attribute set to "Costanza"::

        Q(last_name__eq="Costanza")

    - Match object with ``author.last_name`` attribute set to "Benes"::

        Q(author__last_name__eq="Benes")

    - You are not allowed to specify any key without an operator,
      event though the following would be fine using Django`s Q objects::

        Q(author__last_name="Benes")   # <-- ERROR, will raise ValueError

    - Attributes can be nested arbitrarily deep::

        Q(a__b__c__d__e__f__g__x__gt=3.03)

    - The special ``*__eq=True`` means "match any *true-ish* value"::

        Q(author__account__is_staff__eq=True)

    - Similarly the ``*__eq=False`` means "match any *false-y*" value"::

        Q(author__account__is_disabled=False)

    See :ref:`events-model-filtering-operators`.

    :returns: :class:`collections.Callable`, to match an object with
      the given predicates, call the return value with the object to match:
      ``Q(x__eq==808)(obj)``.

    """

    #: The gate decides the boolean operator of this tree node.
    #: A node can either be *OR* (``a | b``), or an *AND* note (``a & b``).
    #: - Default is *AND*.
    gates = {
        _Q_.OR: any,
        _Q_.AND: all,
    }

    #: If the node is negated (``~a`` / ``a.negate()``), branch will be True,
    #: and we reverse the query into a ``not a`` one.
    branches = {
        True: operator.not_,
        False: operator.truth,
    }

    #: Mapping of opcode to binary operator function: ``f(a, b)``.
    #: Operators may return any true-ish or false-y value.
    operators = {
        'eq': operator.eq,
        'now_eq': wrap_transition(operator.eq, operator.ne),
        'ne': operator.ne,
        'now_ne': wrap_transition(operator.ne, operator.ne),
        'gt': operator.gt,
        'now_gt': wrap_transition(operator.gt, operator.lt),
        'lt': operator.lt,
        'now_lt': wrap_transition(operator.lt, operator.gt),
        'gte': operator.ge,
        'now_gte': wrap_transition(operator.ge, operator.le),
        'lte': operator.le,
        'now_lte': wrap_transition(operator.le, operator.ge),
        'in': reverse_arguments(2)(operator.contains),
        'now_in': wrap_transition(
            reverse_arguments(2)(operator.contains),
            reverse_arguments(2)(not_contains),
        ),
        'not_in': reverse_arguments(2)(not_contains),
        'now_not_in': wrap_transition(
            reverse_arguments(2)(not_contains),
            reverse_arguments(2)(operator.contains),
        ),
        'is': operator.is_,
        'now_is': wrap_transition(operator.is_, operator.is_not),
        'is_not': operator.is_not,
        'now_is_not': wrap_transition(operator.is_not, lambda a, _: a is None),
        'contains': operator.contains,
        'now_contains': wrap_transition(
            operator.contains, negate(operator.contains),
        ),
        'not': lambda x, _: operator.not_(x),
        'true': lambda x, _: operator.truth(x),
        'startswith': startswith,
        'now_startswith': wrap_transition(startswith, negate(startswith)),
        'endswith': endswith,
        'now_endswith': wrap_transition(endswith, negate(endswith)),
    }

    def __call__(self, obj):
        # NOT?( AND|OR(...) )
        return self.branches[self.negated](
            self.gate(f(obj) for f in self.stack)
        )

    def compile(self, fields):
        # this does not traverse the tree, but compiles the nodes
        # in ``self.children`` only.  The nodes below will be compiled
        # and cached when they are called.
        return [self.compile_node(field) for field in fields]

    def compile_node(self, field):
        """Compiles node into a cached function that performs the match.

        :returns: unary :class:`collections.Callable` taking the object
          to match.

        """
        # can embed other Q objects
        if isinstance(field, type(self)):
            return field

        # convert Django Q objects in-place.
        elif isinstance(field, _Q_):
            field.__class__ = type(self)
            return field

        # or it's a key, value pair.
        lhs, rhs = field
        lhs, opcode = self.prepare_statement(lhs, rhs)

        # this creates the new matching function to be added to the stack.
        return self.compile_op(lhs, rhs, opcode)

    def prepare_statement(self, lhs, rhs):
        lhs, _, opcode = lhs.rpartition('__')
        if not opcode or opcode not in self.operators:
            raise ValueError(E_FILTER_FIELD_MISSING_OP.format(lhs))
        return lhs.replace('__', '.'), self.prepare_opcode(opcode, rhs)

    def prepare_opcode(self, O, rhs):
        # eq=True and friends are special, as they should match any
        # true-ish value (__bool__), not check for equality.
        if (O == 'eq' and rhs is True) or O == 'ne' and rhs is False:
            return 'true'
        elif (O == 'eq' and rhs is False) or O == 'ne' and rhs is True:
            return 'not'
        return O

    def compile_op(self, lhs, rhs, opcode):
        return self._compile_op(
            self.apply_trans_op if 'now' in opcode else self.apply_op,
            lhs, rhs, opcode,
        )

    def _compile_op(self, apply, lhs, rhs, opcode, *args):
        return partial(
            apply,
            operator.attrgetter(lhs), self.operators[opcode], rhs, *args
        )

    def apply_op(self, getter, op, rhs, obj, *args):
        # compiled nodes end up being partial versions of this method,
        # with the getter, op and rhs arguments already set.
        return op(getter(obj), rhs, *args)

    def apply_trans_op(self, getter, op, rhs, obj):
        # transition op  (e.g. now_eq) only matches if the
        # value differs from the previous version.
        return self.apply_op(
            getter, op, rhs, obj, self._get_from_prev_version(getter, obj),
        )

    def _get_from_prev_version(self, getter, obj):
        try:
            prev = obj._previous_version
        except AttributeError:
            pass
        else:
            return getter(prev)

    @property
    def gate(self):
        return self.gates[self.connector]

    @cached_property
    def stack(self):
        # the stack is cached on first call.
        return self.compile(self.children)



"""

    thorn.environment
    =================

    Framework integration.

"""
from __future__ import absolute_import, unicode_literals

__all__ = ['envs']


envs = [
    'thorn.environment.django:DjangoEnv',
]



"""

    thorn.django.signals
    ====================

    Django signal dispatchers.

"""
from __future__ import absolute_import, unicode_literals

from operator import attrgetter

from django.core.exceptions import ObjectDoesNotExist

from thorn.generic.signals import signal_dispatcher

from .utils import serialize_model

__all__ = [
    'dispatch_on_create',
    'dispatch_on_change',
    'dispatch_on_delete',
    'dispatch_on_m2m_add',
    'dispatch_on_m2m_remove',
    'dispatch_on_m2m_clear',
]


class dispatch_on_create(signal_dispatcher):

    def setup_signals(self):
        return {'django.db.models.signals.post_save': self}

    def should_dispatch(self, instance, raw=False, created=False, **kwargs):
        return not raw and created


class dispatch_on_change(signal_dispatcher):

    def setup_signals(self):
        return {
            'django.db.models.signals.pre_save': self.on_pre_save,
            'django.db.models.signals.post_save': self,
        }

    def on_pre_save(self, instance, sender, raw=False, **kwargs):
        if self.use_transitions and not raw and instance.pk:
            try:
                instance._previous_version = sender.objects.get(pk=instance.pk)
            except ObjectDoesNotExist:
                pass

    def should_dispatch(self, instance, created=False, raw=False, **kwargs):
        return not raw and not created


class dispatch_on_delete(signal_dispatcher):

    def setup_signals(self):
        return {'django.db.models.signals.post_delete': self}


class dispatch_on_m2m_change(signal_dispatcher):

    def __init__(self, fun, related_field, **kwargs):
        super(dispatch_on_m2m_change, self).__init__(fun, **kwargs)
        self.related_field = related_field
        self.actions = self.setup_actions()

    def setup_actions(self):
        return {}

    def setup_signals(self):
        return {'django.db.models.signals.m2m_changed': self.on_m2m_change}

    def prepare_sender(self, sender):
        return attrgetter(self.related_field)(sender).through

    def context(self, instance, model, pk_set, **kwargs):
        return {
            'instance': instance.pk,
            'model': serialize_model(model),
            'pk_set': list(sorted(pk_set)),
        }

    def on_m2m_change(self, sender, action, instance, model, **kwargs):
        try:
            handler = self.actions[action]
        except KeyError:
            pass
        else:
            handler(instance, sender=sender, model=model, **kwargs)


class dispatch_on_m2m_add(dispatch_on_m2m_change):

    def setup_actions(self):
        return {'post_add': self}


class dispatch_on_m2m_remove(dispatch_on_m2m_change):

    def setup_actions(self):
        return {'post_remove': self}


class dispatch_on_m2m_clear(dispatch_on_m2m_change):

    def setup_actions(self):
        return {'post_clear': self}



from __future__ import absolute_import, unicode_literals

from thorn.django.models import Subscriber

from kombu.five import text_t

from django.contrib.auth import get_user_model

from thorn.tests.case import Case


class test_Subscriber(Case):

    def setUp(self):
        self.user = get_user_model()(username='george')
        self.subscriber = Subscriber(
            event='foo.created', user=self.user, url='http://example.com',
        )

    def test_str(self):
        self.assertTrue(text_t(self.subscriber))

    def test_as_dict(self):
        self.assertDictEqual(self.subscriber.as_dict(), {
            'event': self.subscriber.event,
            'user': self.subscriber.user.pk,
            'url': self.subscriber.url,
            'content_type': self.subscriber.content_type,
            'hmac_secret': self.subscriber.hmac_secret,
            'hmac_digest': self.subscriber.hmac_digest,
        })

    def test_from_dict__arg(self):
        x = Subscriber.from_dict('http://e.com', event='foo.bar')
        self.assertIsInstance(x, Subscriber)
        self.assertEqual(x.url, 'http://e.com')
        self.assertEqual(x.event, 'foo.bar')

    def test_from_dict__dict(self):
        x = Subscriber.from_dict(url='http://e.com', event='foo.bar')
        self.assertIsInstance(x, Subscriber)
        self.assertEqual(x.url, 'http://e.com')
        self.assertEqual(x.event, 'foo.bar')



"""

    thorn.django.rest_framework.views
    =================================

    API endpoints for users to create and manage their webhook subscriptions.

"""
from __future__ import absolute_import, unicode_literals

from rest_framework.generics import (
    ListCreateAPIView, RetrieveUpdateDestroyAPIView,
)
from rest_framework.permissions import IsAuthenticated

from django.shortcuts import get_object_or_404

from thorn._state import current_app
from thorn.django.models import Subscriber

from .serializers import SubscriberSerializer

__all__ = ['SubscriberList', 'SubscriberDetail']


def _permission_classes(view):
    perms = current_app().settings.THORN_DRF_PERMISSION_CLASSES
    return perms if perms is not None else (IsAuthenticated,)


class SubscriberList(ListCreateAPIView):
    """List and create new subscriptions for the currently logged in user."""

    serializer_class = SubscriberSerializer
    model = Subscriber
    permission_classes = property(_permission_classes)

    def get_queryset(self):
        return self.model.objects.filter(user=self.request.user)

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class SubscriberDetail(RetrieveUpdateDestroyAPIView):
    """Update, delete or get details for specific subscription owned by
    the currently logged in user."""

    serializer_class = SubscriberSerializer
    model = Subscriber
    lookup_field = 'uuid'
    permission_classes = property(_permission_classes)

    def get_object(self):
        return get_object_or_404(
            self.model, user=self.request.user, uuid=self.kwargs['uuid'],
        )



from __future__ import absolute_import, unicode_literals

from .case import Case


class test_modules(Case):

    def test_admin(self):
        import thorn.django.admin  # noqa

    def test_apps(self):
        import thorn.django.apps  # noqa

    def test_rest_framework_urls(self):
        import thorn.django.rest_framework.urls  # noqa



"""

    thorn.django.utils
    ==================

    Django-related utilities.

"""
from __future__ import absolute_import, unicode_literals

from collections import Mapping

from django.apps import apps

__all__ = ['serialize_model', 'deserialize_model']


def serialize_model(m):
    return {'app_label': m._meta.app_label, 'model_name': m._meta.model_name}


def deserialize_model(m):
    if isinstance(m, Mapping):
        return apps.get_model(m['app_label'], m['model_name'])
    return m



"""

    thorn.models
    ============

    Generic base model mixins.

"""
from __future__ import absolute_import, unicode_literals

import hashlib

from abc import ABCMeta, abstractmethod, abstractproperty
from six import string_types

from celery.five import with_metaclass
from itsdangerous import Signer

__all__ = ['AbstractSubscriber', 'SubscriberModelMixin', 'get_digest']


def get_digest(d):
    assert d in hashlib.algorithms_available
    return getattr(hashlib, d)


@with_metaclass(ABCMeta)
class AbstractSubscriber(object):

    #: Unique identifier.
    uuid = abstractproperty()

    #: Event pattern this subscriber is subscribed to (e.g. ``article.*``).
    event = abstractproperty()

    #: Destination URL to dispatch this event.
    url = abstractproperty()

    #: User filter: when set only dispatch if the event sender matches.
    user = abstractproperty()

    #: HMAC secret key, of arbitrary length.
    hmac_secret = abstractproperty()

    #: HMAC digest type (e.g. ``"sha512"``).
    #:
    #: The value used must be a member of :data:`hashlib.algorithms_available`.
    hmac_digest = abstractproperty()

    #: MIME-type to use for web requests made to the subscriber :attr:`url`.
    content_type = abstractproperty()

    @abstractmethod
    def as_dict(self):
        """Dictionary representation of Subscriber."""
        pass  # pragma: no cover

    @abstractmethod
    def from_dict(self, *args, **kwargs):
        """Create subscriber from dictionary representation.

        Accept the same arguments as :class:`dict`.

        """
        pass  # pragma: no cover

    @abstractmethod
    def sign(self, message):
        """Sign (HMAC) message using :attr:`hmac_secret` and the current
        :attr:`hmac_digest` type."""
        pass  # pragma: no cover

    @abstractmethod
    def user_ident(self):
        """Return :attr:`user` identity, must be json serializable like a
        database primary key."""
        pass  # pragma: no cover

    @classmethod
    def register(cls, other):
        # we override `register` to return other for use as a decorator.
        type(cls).register(cls, other)
        return other


@AbstractSubscriber.register
class SubscriberModelMixin(object):

    @classmethod
    def from_dict(cls, *args, **kwargs):
        if args and isinstance(args[0], string_types):
            args = ({'url': args[0]},)
        return cls(**dict(*args, **kwargs))

    def as_dict(self):
        return {
            'event': self.event,
            'user': self.user_ident(),
            'url': self.url,
            'hmac_secret': self.hmac_secret,
            'hmac_digest': self.hmac_digest,
            'content_type': self.content_type,
        }

    def sign(self, message):
        return Signer(
            self.hmac_secret,
            digest_method=get_digest(self.hmac_digest)).get_signature(message)



from __future__ import absolute_import, unicode_literals

from thorn.conf import Settings, event_choices, all_settings
from thorn.exceptions import ImproperlyConfigured

from .case import Case, Mock


class test_Settings(Case):

    attributes = [
        ('THORN_CHUNKSIZE', 'default_chunksize'),
        ('THORN_CODECS', 'default_codecs'),
        ('THORN_DISPATCHER', 'default_dispatcher'),
        ('THORN_DRF_PERMISSION_CLASSES', 'default_drf_permission_classes'),
        ('THORN_EVENT_CHOICES', 'default_event_choices'),
        ('THORN_EVENT_TIMEOUT', 'default_timeout'),
    ]

    def setup(self):
        self.app = Mock(name='app')

    def test_settings(self):
        for key_attr, default_attr in self.attributes:
            s1 = Settings(app=self.app)
            setattr(self.app.config, key_attr, None)
            self.assertEqual(
                getattr(s1, key_attr), getattr(s1, default_attr), key_attr)

            setattr(self.app.config, key_attr, 'just')
            s2 = Settings(app=self.app)
            self.assertEqual(getattr(s2, key_attr), 'just', key_attr)

    def test_THORN_SUBSCRIBERS(self):
        self.app.config.THORN_SUBSCRIBERS = None
        self.assertDictEqual(Settings(app=self.app).THORN_SUBSCRIBERS, {})
        self.app.config.THORN_SUBSCRIBERS = 'just'
        self.assertEqual(Settings(app=self.app).THORN_SUBSCRIBERS, 'just')


class test_event_choices(Case):

    def setup(self):
        self.app = Mock(name='app')

    def test_wrong_type(self):
        self.app.settings.THORN_EVENT_CHOICES = 3
        with self.assertRaises(ImproperlyConfigured):
            event_choices(app=self.app)

    def test(self):
        self.app.settings.THORN_EVENT_CHOICES = ('a.b', 'c.d', 'e.f')
        self.assertListEqual(
            event_choices(app=self.app),
            [('a.b', 'a.b'), ('c.d', 'c.d'), ('e.f', 'e.f')],
        )


class test_all_settings(Case):

    def test(self):
        self.assertTrue(all_settings())



"""

    thorn.django.models
    ===================

    Django models required to dispatch webhook events.

"""
from __future__ import absolute_import, unicode_literals

from uuid import uuid4

from django.conf import settings as django_settings
from django.db import models
from django.utils import text
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from thorn.conf import event_choices, MIME_JSON, MIME_URLFORM
from thorn.generic.models import AbstractSubscriber, SubscriberModelMixin
from thorn.utils.compat import random_secret

from .managers import SubscriberManager

__all__ = ['Subscriber']

#: max size when indexing InnoDB utf8mb4
#: reportedly this is caused by MySQL being purposefully and irreversibly
#: brain-damaged: https://code.djangoproject.com/ticket/18392
CHAR_MAX_LENGTH = 190

CONTENT_TYPES = {
    MIME_JSON,
    MIME_URLFORM,
}


# Django migrations cannot handle partial objects, ugh...
def random_secret64():
    return random_secret(64)


@AbstractSubscriber.register
@python_2_unicode_compatible
class Subscriber(models.Model, SubscriberModelMixin):
    objects = SubscriberManager()

    uuid = models.UUIDField(
        _('UUID'),
        default=uuid4, editable=False, unique=True, null=False,
        help_text=_('Unique identifier for this subscriber.')
    )

    event = models.CharField(
        _('event'),
        max_length=CHAR_MAX_LENGTH,
        choices=event_choices(),
        db_index=True,
        help_text=_('Name of event to connect with'),
    )

    url = models.URLField(
        _('URL'),
        max_length=CHAR_MAX_LENGTH,
        db_index=True,
        help_text=_('Callback URL'),
    )

    user = models.ForeignKey(
        django_settings.AUTH_USER_MODEL,
        related_name='%(app_label)s_%(class)s',
        null=True,
    )

    hmac_secret = models.TextField(
        _('HMAC Secret'),
        default=random_secret64,
        help_text=_('Specify HMAC secret for endpoints to verify'),
    )

    hmac_digest = models.CharField(
        _('HMAC Digest Type'),
        max_length=64,
        choices=[
            ('SHA512', 'sha512'),
            ('SHA256', 'sha256'),
            ('SHA1', 'sha1'),
        ],
        default='sha256',
        help_text=_('Specify HMAC digest type (use sha256 if uncertain)'),
    )

    content_type = models.CharField(
        _('content type'),
        max_length=CHAR_MAX_LENGTH,
        choices=zip(CONTENT_TYPES, CONTENT_TYPES),
        default=MIME_JSON,
        help_text='Desired content type for requests to this callback.'
    )

    created_at = models.DateTimeField(
        _('created at'), editable=False, auto_now_add=True)

    updated_at = models.DateTimeField(
        _('updated_at'), editable=False, auto_now=True,
    )

    class Meta:
        verbose_name = _('subscriber')
        verbose_name_plural = _('subscriber')
        # ordering by hostname for ability to optimize for keepalive.
        ordering = ['url', '-created_at']
        get_latest_by = 'updated_at'
        unique_together = ('url', 'event')

    def user_ident(self):
        return self.user and self.user.pk

    def __str__(self):
        return '{0} -> {1}'.format(
            self.event, text.Truncator(self.url).chars(43),
        )



"""

    thorn.dispatch.celery
    =====================

    Celery-based webhook dispatcher.

"""
from __future__ import absolute_import, unicode_literals

from celery import group

from thorn.tasks import send_event, dispatch_requests
from thorn.utils.functional import groupbymax

from . import base

__all__ = ['Dispatcher', 'WorkerDispatcher']


class Dispatcher(base.Dispatcher):
    """Dispatcher using Celery tasks to dispatch events.

    Overrides what happens when :meth:`thorn.webhook.Event.send` is called
    so that dispatching the HTTP request tasks is performed by a worker,
    instead of in the current process.

    """

    def send(self, event, payload, sender, timeout=None, **kwargs):
        return send_event.s(
            event, payload,
            sender.pk if sender else sender, timeout,
        ).apply_async()


class WorkerDispatcher(base.Dispatcher):
    """Dispatcher used by the :func:`thorn.tasks.send_event` task."""

    def send(self, event, payload, sender, timeout=None, **kwargs):
        # the requests are sorted by url, so we group them into chunks
        # each containing a list of requests for that host/port/scheme pair,
        # with up to :setting:`THORN_CHUNKSIZE` requests each.
        #
        # this way requests have a good chance of reusing keepalive
        # connections as requests with the same host are grouped together.
        return group(
            dispatch_requests.s([req.as_dict() for req in chunk])
            for chunk in self.group_requests(
                self.prepare_requests(event, payload, sender, timeout))
        ).delay()

    def group_requests(self, requests):
        """Group requests by keep-alive host/port/scheme ident."""
        return groupbymax(
            requests,
            max=self.app.settings.THORN_CHUNKSIZE,
            key=self._compare_requests,
        )

    def _compare_requests(self, a, b):
        return a.urlident == b.urlident



"""

    thorn.django.apps
    =================

    Django application configurations.

"""
from __future__ import absolute_import, unicode_literals

from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _

__all__ = ['WebhooksConfig']


class WebhooksConfig(AppConfig):
    name = 'thorn.django'
    label = 'webhooks'
    verbose_name = _('Webhooks')



from __future__ import absolute_import, unicode_literals

from thorn.generic.models import SubscriberModelMixin, get_digest

from thorn.tests.case import Case, Mock, patch


class DigestCase(Case):

    def setup(self):
        self.hashlib = self.patch('thorn.generic.models.hashlib')
        self.hashlib.algorithms_available = ['sha1']


class test_get_digest(DigestCase):

    def test_get_digest(self):
        self.assertIs(get_digest('sha1'), self.hashlib.sha1)

    def test_get_digest__algorithm_unavailable(self):
        with self.assertRaises(AssertionError):
            get_digest('sha256')


class test_SubscriberModelMixin(DigestCase):

    @patch('thorn.generic.models.Signer')
    def test_sign(self, Signer):
        x = SubscriberModelMixin()
        x.hmac_secret = Mock(name='hmac_secret')
        x.hmac_digest = 'sha1'
        res = x.sign('thequickbrownfox')
        Signer.assert_called_with(
            x.hmac_secret, digest_method=self.hashlib.sha1,
        )
        Signer().get_signature.assert_called_with('thequickbrownfox')
        self.assertIs(res, Signer().get_signature())
